from collections import OrderedDict
import hub
import pytest
import numpy as np
from hub.util.diff import (
    get_all_changes_string,
    get_lowest_common_ancestor,
    sanitize_commit,
)
from hub.util.remove_cache import get_base_storage
from hub.util.exceptions import (
    CheckoutError,
    CommitError,
    ReadOnlyModeError,
    InfoError,
    TensorModifiedError,
    EmptyCommitError,
)

NO_COMMIT_PASSED_DIFF = ""
ONE_COMMIT_PASSED_DIFF = "The 2 diffs are calculated relative to the most recent common ancestor (%s) of the current state and the commit passed."
TWO_COMMIT_PASSED_DIFF = "The 2 diffs are calculated relative to the most recent common ancestor (%s) of the two commits passed."


def get_lca_id_helper(version_state, id_1, id_2=None):
    id_1 = sanitize_commit(id_1, version_state)
    id_2 = sanitize_commit(id_2, version_state) if id_2 else version_state["commit_id"]
    commit_node_1 = version_state["commit_node_map"][id_1]
    commit_node_2 = version_state["commit_node_map"][id_2]
    return get_lowest_common_ancestor(commit_node_1, commit_node_2)


def commit_details_helper(commits, ds):
    for commit in commits:
        assert ds.get_commit_details(commit["commit"]) == commit


def get_diff_helper(
    ds_changes_1,
    ds_changes_2,
    tensor_changes_1,
    tensor_changes_2,
    version_state=None,
    a=None,
    b=None,
):
    if a and b:
        lca_id = get_lca_id_helper(version_state, a, b)
        message0 = TWO_COMMIT_PASSED_DIFF % lca_id
        message1 = f"Diff in {a} (target id 1):\n"
        message2 = f"Diff in {b} (target id 2):\n"
    elif a:
        lca_id = get_lca_id_helper(version_state, a)
        message0 = ONE_COMMIT_PASSED_DIFF % lca_id
        message1 = "Diff in HEAD:\n"
        message2 = f"Diff in {a} (target id):\n"
    else:
        message0 = NO_COMMIT_PASSED_DIFF
        message1 = "Diff in HEAD relative to the previous commit:\n"
        message2 = ""

    target = (
        get_all_changes_string(
            ds_changes_1,
            ds_changes_2,
            tensor_changes_1,
            tensor_changes_2,
            message0,
            message1,
            message2,
        )
        + "\n"
    )

    return target


def tensor_diff_helper(
    data_added=[0, 0],
    data_updated=set(),
    created=False,
    cleared=False,
    info_updated=False,
    data_transformed_in_place=False,
    data_deleted=set(),
):
    return {
        "data_added": data_added,
        "data_updated": data_updated,
        "created": created,
        "cleared": cleared,
        "info_updated": info_updated,
        "data_transformed_in_place": data_transformed_in_place,
        "data_deleted": data_deleted,
    }


def test_commit(local_ds):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append(1)
        local_ds.log()
        a = local_ds.commit("first")
        local_ds.abc[0] = 2
        b = local_ds.commit("second")
        local_ds.abc[0] = 3
        c = local_ds.commit("third")
        assert local_ds.abc[0].numpy() == 3
        local_ds.checkout(a)
        assert local_ds.commit_id == a
        assert local_ds.abc[0].numpy() == 1
        local_ds.checkout(b)
        assert local_ds.commit_id == b
        assert local_ds.abc[0].numpy() == 2
        local_ds.checkout(c)
        assert local_ds.commit_id == c
        assert local_ds.branch == "main"
        assert local_ds.abc[0].numpy() == 3
        with pytest.raises(CheckoutError):
            local_ds.checkout("main", create=True)
        with pytest.raises(CheckoutError):
            local_ds.checkout(a, create=True)


"""
test for checking unchanged dataset commits
"""


def test_unchanged_commit(local_ds):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append(1)
        local_ds.log()
        a = local_ds.commit("first")
        local_ds.checkout(a)
        assert local_ds.commit_id == a
        with pytest.raises(EmptyCommitError):
            b = local_ds.commit("second")
        c = local_ds.commit("third", allow_empty=True)
        local_ds.checkout(c)
        assert local_ds.commit_id == c


def test_commit_checkout(local_ds):
    with local_ds:
        local_ds.create_tensor("img")
        local_ds.img.extend(np.ones((10, 100, 100, 3)))
        first_commit_id = local_ds.commit("stored all ones")

        for i in range(5):
            local_ds.img[i] *= 2
        second_commit_id = local_ds.commit("multiplied value of some images by 2")

        for i in range(5):
            assert (local_ds.img[i].numpy() == 2 * np.ones((100, 100, 3))).all()
        local_ds.checkout(first_commit_id)  # now all images are ones again

        for i in range(10):
            assert (local_ds.img[i].numpy() == np.ones((100, 100, 3))).all()

        local_ds.checkout("alternate", create=True)
        assert local_ds.branch == "alternate"

        for i in range(5):
            local_ds.img[i] *= 3
        local_ds.commit("multiplied value of some images by 3")

        for i in range(5):
            assert (local_ds.img[i].numpy() == 3 * np.ones((100, 100, 3))).all()

        local_ds.checkout(second_commit_id)  # first 5 images are 2s, rest are 1s now
        assert local_ds.commit_id == second_commit_id
        assert local_ds.branch == "main"

        # we are not at the head of master but rather at the last commit, so we automatically get checked out to a new branch here
        for i in range(5, 10):
            local_ds.img[i] *= 2
        local_ds.commit("multiplied value of remaining images by 2")

        for i in range(10):
            assert (local_ds.img[i].numpy() == 2 * np.ones((100, 100, 3))).all()

        local_ds.checkout("alternate")

        for i in range(5, 10):
            local_ds.img[i] *= 3

        for i in range(10):
            assert (local_ds.img[i].numpy() == 3 * np.ones((100, 100, 3))).all()
        local_ds.commit("multiplied value of remaining images by 3")
        for i in range(10):
            assert (local_ds.img[i].numpy() == 3 * np.ones((100, 100, 3))).all()


def test_commit_checkout_2(local_ds):
    local_ds.create_tensor("abc")
    local_ds.create_tensor("img")
    for i in range(10):
        local_ds.img.append(i * np.ones((100, 100, 3)))
    a = local_ds.commit("first")

    local_ds.img[7] *= 2

    assert (local_ds.img[6].numpy() == 6 * np.ones((100, 100, 3))).all()
    assert (local_ds.img[7].numpy() == 2 * 7 * np.ones((100, 100, 3))).all()
    assert (local_ds.img[8].numpy() == 8 * np.ones((100, 100, 3))).all()
    assert (local_ds.img[9].numpy() == 9 * np.ones((100, 100, 3))).all()

    assert (local_ds.img[2].numpy() == 2 * np.ones((100, 100, 3))).all()

    b = local_ds.commit("second")

    # going back to first commit
    local_ds.checkout(a)

    assert (local_ds.img[7].numpy() == 7 * np.ones((100, 100, 3))).all()

    local_ds.checkout("another", create=True)

    local_ds.img[7] *= 3

    # and not 6 * 7 as it would have been, had we checked out from b
    assert (local_ds.img[7].numpy() == 3 * 7 * np.ones((100, 100, 3))).all()

    local_ds.commit("first2")

    local_ds.checkout("main")
    assert (local_ds.img[7].numpy() == 2 * 7 * np.ones((100, 100, 3))).all()
    local_ds.log()


def test_auto_checkout_bug(local_ds):
    local_ds.create_tensor("abc")
    local_ds.abc.extend([1, 2, 3, 4, 5])
    a = local_ds.commit("it is 1")
    local_ds.abc[0] = 2
    b = local_ds.commit("it is 2")
    c = local_ds.checkout(a)
    local_ds.checkout("other", True)
    d = local_ds.pending_commit_id
    local_ds.abc[0] = 3
    e = local_ds.commit("it is 3")
    local_ds.checkout(b)
    local_ds.abc[0] = 4
    f = local_ds.commit("it is 4")
    g = local_ds.checkout(a)
    local_ds.abc[0] = 5
    dsv = local_ds[0:3]
    dsv.abc[0] = 5
    h = local_ds.commit("it is 5")
    i = local_ds.checkout(e)
    local_ds.abc[0] = 6
    tsv = local_ds.abc[0:5]
    tsv[0] = 6
    j = local_ds.commit("it is 6")
    local_ds.log()
    local_ds.checkout(a)
    assert dsv.abc[0].numpy() == 1
    assert local_ds.abc[0].numpy() == 1
    local_ds.checkout(b)
    assert local_ds.abc[0].numpy() == 2
    local_ds.checkout(c)
    assert local_ds.abc[0].numpy() == 1
    local_ds.checkout(d)
    assert local_ds.abc[0].numpy() == 3
    local_ds.checkout(e)
    assert local_ds.abc[0].numpy() == 3
    local_ds.checkout(f)
    assert local_ds.abc[0].numpy() == 4
    local_ds.checkout(g)
    assert local_ds.abc[0].numpy() == 1
    local_ds.checkout(h)
    assert local_ds.abc[0].numpy() == 5
    local_ds.checkout(i)
    assert local_ds.abc[0].numpy() == 3
    local_ds.checkout(j)
    assert local_ds.abc[0].numpy() == 6
    local_ds.checkout("main")
    assert local_ds.abc[0].numpy() == 2
    local_ds.abc[0] = 7
    local_ds.checkout("copy", True)
    assert local_ds.abc[0].numpy() == 7
    local_ds.checkout("other")
    assert local_ds.abc[0].numpy() == 3


def test_read_mode(local_ds):
    base_storage = get_base_storage(local_ds.storage)
    base_storage.enable_readonly()
    ds = hub.Dataset(storage=local_ds.storage, read_only=True, verbose=False)
    with pytest.raises(ReadOnlyModeError):
        ds.commit("first")
    with pytest.raises(ReadOnlyModeError):
        ds.checkout("third", create=True)


def test_checkout_address_not_found(local_ds):
    with pytest.raises(CheckoutError):
        local_ds.checkout("second")


def test_dynamic(local_ds):
    local_ds.create_tensor("img")
    for i in range(10):
        local_ds.img.append(i * np.ones((100, 100, 3)))

    a = local_ds.commit("first")
    for i in range(10):
        local_ds.img[i] = 2 * i * np.ones((150, 150, 3))
    local_ds.checkout(a)

    for i in range(10):
        assert (local_ds.img[i].numpy() == i * np.ones((100, 100, 3))).all()

    local_ds.checkout("main")

    for i in range(10):
        assert (local_ds.img[i].numpy() == 2 * i * np.ones((150, 150, 3))).all()


def test_different_lengths(local_ds):
    with local_ds:
        local_ds.create_tensor("img")
        local_ds.create_tensor("abc")
        local_ds.img.extend(np.ones((5, 50, 50)))
        local_ds.abc.extend(np.ones((2, 10, 10)))
        first = local_ds.commit("stored 5 images, 2 abc")
        local_ds.img.extend(np.ones((3, 50, 50)))
        second = local_ds.commit("stored 3 more images")
        assert len(local_ds.tensors) == 2
        assert len(local_ds.img) == 8
        assert (local_ds.img.numpy() == np.ones((8, 50, 50))).all()
        assert len(local_ds.abc) == 2
        assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()
        local_ds.checkout(first)
        assert len(local_ds.tensors) == 2
        assert len(local_ds.img) == 5
        assert (local_ds.img.numpy() == np.ones((5, 50, 50))).all()
        assert len(local_ds.abc) == 2
        assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()

        # will autocheckout to new branch
        local_ds.create_tensor("ghi")
        local_ds.ghi.extend(np.ones((2, 10, 10)))
        local_ds.img.extend(np.ones((2, 50, 50)))
        local_ds.abc.extend(np.ones((3, 10, 10)))
        assert len(local_ds.tensors) == 3
        assert len(local_ds.img) == 7
        assert (local_ds.img.numpy() == np.ones((7, 50, 50))).all()
        assert len(local_ds.abc) == 5
        assert (local_ds.abc.numpy() == np.ones((5, 10, 10))).all()
        assert len(local_ds.ghi) == 2
        assert (local_ds.ghi.numpy() == np.ones((2, 10, 10))).all()
        third = local_ds.commit(
            "stored 2 more images, 3 more abc in other branch, created ghi"
        )
        local_ds.checkout(first)
        assert len(local_ds.tensors) == 2
        assert len(local_ds.img) == 5
        assert (local_ds.img.numpy() == np.ones((5, 50, 50))).all()
        assert len(local_ds.abc) == 2
        assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()
        local_ds.checkout(second)
        assert len(local_ds.tensors) == 2
        assert len(local_ds.img) == 8
        assert (local_ds.img.numpy() == np.ones((8, 50, 50))).all()
        assert len(local_ds.abc) == 2
        assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()
        local_ds.checkout(third)
        assert len(local_ds.tensors) == 3
        assert len(local_ds.img) == 7
        assert (local_ds.img.numpy() == np.ones((7, 50, 50))).all()
        assert len(local_ds.abc) == 5
        assert (local_ds.abc.numpy() == np.ones((5, 10, 10))).all()
        local_ds.checkout("main")
        assert len(local_ds.tensors) == 2
        assert len(local_ds.img) == 8
        assert (local_ds.img.numpy() == np.ones((8, 50, 50))).all()
        assert len(local_ds.abc) == 2
        assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()

    path = local_ds.path
    if path.startswith("mem://"):
        # memory datasets are not persistent
        return

    # reloading the dataset to check persistence

    local_ds = hub.dataset(path)
    assert len(local_ds.tensors) == 2
    assert len(local_ds.img) == 8
    assert (local_ds.img.numpy() == np.ones((8, 50, 50))).all()
    assert len(local_ds.abc) == 2
    assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()
    local_ds.checkout(first)
    assert len(local_ds.tensors) == 2
    assert len(local_ds.img) == 5
    assert (local_ds.img.numpy() == np.ones((5, 50, 50))).all()
    assert len(local_ds.abc) == 2
    assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()
    local_ds.checkout(second)
    assert len(local_ds.tensors) == 2
    assert len(local_ds.img) == 8
    assert (local_ds.img.numpy() == np.ones((8, 50, 50))).all()
    assert len(local_ds.abc) == 2
    assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()
    local_ds.checkout(third)
    assert len(local_ds.tensors) == 3
    assert len(local_ds.img) == 7
    assert (local_ds.img.numpy() == np.ones((7, 50, 50))).all()
    assert len(local_ds.abc) == 5
    assert (local_ds.abc.numpy() == np.ones((5, 10, 10))).all()
    local_ds.checkout("main")
    assert len(local_ds.tensors) == 2
    assert len(local_ds.img) == 8
    assert (local_ds.img.numpy() == np.ones((8, 50, 50))).all()
    assert len(local_ds.abc) == 2
    assert (local_ds.abc.numpy() == np.ones((2, 10, 10))).all()


def test_auto_checkout(local_ds):
    # auto checkout happens when write operations are performed on non head commits
    local_ds.create_tensor("abc")
    first = local_ds.commit("created abc")

    local_ds.checkout(first)
    assert local_ds.branch == "main"
    local_ds.create_tensor("def")
    assert local_ds.branch != "main"

    local_ds.checkout(first)
    assert local_ds.branch == "main"
    local_ds.abc.append(1)
    assert local_ds.branch != "main"

    local_ds.checkout(first)
    assert local_ds.branch == "main"
    local_ds.abc.extend([1])
    assert local_ds.branch != "main"

    local_ds.checkout(first)
    assert local_ds.branch == "main"

    with pytest.raises(InfoError):
        local_ds.info[5] = 5

    assert local_ds.branch == "main"


def test_auto_commit(local_ds):
    initial_commit_id = local_ds.pending_commit_id
    # auto commit as head of main branch and has changes (due to creation of dataset)
    local_ds.checkout("pqr", create=True)
    local_ds.checkout("main")
    second_commit_id = local_ds.pending_commit_id
    assert second_commit_id != initial_commit_id
    assert local_ds.commit_id == initial_commit_id
    local_ds.create_tensor("abc")
    local_ds.abc.append(1)
    # auto commit as head of main and data present
    local_ds.checkout("xyz", create=True)
    local_ds.checkout("main")

    assert local_ds.pending_commit_id != second_commit_id
    assert local_ds.commit_id == second_commit_id

    with local_ds:
        local_ds.abc.append(1)

    third_commit_id = local_ds.pending_commit_id

    # auto commit as head of main and data present
    local_ds.checkout("tuv", create=True)
    local_ds.checkout("main")

    assert local_ds.pending_commit_id != third_commit_id
    assert local_ds.commit_id == third_commit_id


def test_dataset_info(local_ds):
    assert len(local_ds.info) == 0
    local_ds.info.key = "value"
    assert len(local_ds.info) == 1
    assert local_ds.info.key == "value"

    a = local_ds.commit("added key, value", allow_empty=True)
    assert len(local_ds.info) == 1
    assert local_ds.info.key == "value"

    local_ds.info.key2 = "value2"
    assert len(local_ds.info) == 2
    assert local_ds.info.key == "value"
    assert local_ds.info.key2 == "value2"

    b = local_ds.commit("added key2, value2", allow_empty=True)
    assert len(local_ds.info) == 2
    assert local_ds.info.key == "value"
    assert local_ds.info.key2 == "value2"

    local_ds.checkout(a)
    assert local_ds.info.key == "value"

    local_ds.checkout("alt", create=True)
    local_ds.info.key = "notvalue"
    assert len(local_ds.info) == 1
    assert local_ds.info.key == "notvalue"
    c = local_ds.commit("changed key to notvalue", allow_empty=True)

    local_ds.checkout(a)
    assert len(local_ds.info) == 1
    assert local_ds.info.key == "value"

    local_ds.checkout(b)
    assert len(local_ds.info) == 2
    assert local_ds.info.key == "value"
    assert local_ds.info.key2 == "value2"

    local_ds.checkout("alt")
    assert len(local_ds.info) == 1
    assert local_ds.info.key == "notvalue"

    local_ds.checkout(c)
    assert len(local_ds.info) == 1
    assert local_ds.info.key == "notvalue"


def test_tensor_info(local_ds):
    local_ds.create_tensor("abc")
    assert len(local_ds.abc.info) == 0
    local_ds.abc.info.key = "value"
    assert len(local_ds.abc.info) == 1
    assert local_ds.abc.info.key == "value"

    a = local_ds.commit("added key, value")
    assert len(local_ds.abc.info) == 1
    assert local_ds.abc.info.key == "value"

    local_ds.abc.info.key2 = "value2"
    assert len(local_ds.abc.info) == 2
    assert local_ds.abc.info.key == "value"
    assert local_ds.abc.info.key2 == "value2"

    b = local_ds.commit("added key2, value2")
    assert len(local_ds.abc.info) == 2
    assert local_ds.abc.info.key == "value"
    assert local_ds.abc.info.key2 == "value2"

    local_ds.checkout(a)
    assert local_ds.abc.info.key == "value"

    local_ds.checkout("alt", create=True)
    local_ds.abc.info.key = "notvalue"
    assert len(local_ds.abc.info) == 1
    assert local_ds.abc.info.key == "notvalue"
    c = local_ds.commit("changed key to notvalue")

    local_ds.checkout(a)
    assert len(local_ds.abc.info) == 1
    assert local_ds.abc.info.key == "value"

    local_ds.checkout(b)
    assert len(local_ds.abc.info) == 2
    assert local_ds.abc.info.key == "value"
    assert local_ds.abc.info.key2 == "value2"

    local_ds.checkout("alt")
    assert len(local_ds.abc.info) == 1
    assert local_ds.abc.info.key == "notvalue"

    local_ds.checkout(c)
    assert len(local_ds.abc.info) == 1
    assert local_ds.abc.info.key == "notvalue"


def test_delete(local_ds):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append(1)
        a = local_ds.commit("first")
        local_ds.delete_tensor("abc")
        b = local_ds.commit("second", allow_empty=True)
        local_ds.checkout(a)
        assert local_ds.abc[0].numpy() == 1
        local_ds.checkout(b)
        assert local_ds.tensors == {}

        local_ds.create_tensor("x/y/z")
        local_ds["x/y/z"].append(1)
        c = local_ds.commit("third")
        local_ds["x"].delete_tensor("y/z")
        d = local_ds.commit("fourth", allow_empty=True)
        local_ds.checkout(c)
        assert local_ds["x/y/z"][0].numpy() == 1
        local_ds.checkout(d)
        assert local_ds.tensors == {}
        assert list(local_ds.groups) == ["x"]
        local_ds.delete_group("x")
        assert list(local_ds.groups) == []

        local_ds.checkout(c)
        local_ds["x"].delete_group("y")
        assert local_ds.tensors == {}
        assert list(local_ds.groups) == ["x"]

        local_ds.checkout(c)
        local_ds.delete_group("x/y")
        assert local_ds.tensors == {}
        assert list(local_ds.groups) == ["x"]


def test_tensor_rename(local_ds):
    with local_ds:
        local_ds.create_tensor("x/y/z")
        local_ds["x/y/z"].append(1)
        local_ds["x/y"].rename_tensor("z", "a")
        a = local_ds.commit("first")

        assert local_ds["x/y/a"][0].numpy() == 1
        local_ds["x/y/a"].append(2)
        local_ds["x"].rename_tensor("y/a", "y/z")
        b = local_ds.commit("second")

        assert local_ds["x/y/z"][1].numpy() == 2
        local_ds.create_tensor("x/y/a")
        local_ds["x/y/a"].append(3)
        local_ds["x/y"].rename_tensor("z", "b")
        c = local_ds.commit("third")

        local_ds.checkout(a)
        assert local_ds["x/y/a"][0].numpy() == 1

        local_ds.checkout(b)
        assert local_ds["x/y/z"][1].numpy() == 2

        local_ds.checkout(c)
        assert local_ds["x/y/a"][0].numpy() == 3
        assert local_ds["x/y/b"][1].numpy() == 2


def test_dataset_diff(local_ds, capsys):
    local_ds.create_tensor("abc")
    a = local_ds.commit()
    local_ds.rename_tensor("abc", "xyz")
    local_ds.info["hello"] = "world"

    local_ds.diff()
    dataset_changes = {"info_updated": True, "renamed": {"abc": "xyz"}}
    target = get_diff_helper(dataset_changes, {}, {}, None)
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == {}, {}

    b = local_ds.commit()
    local_ds.delete_tensor("xyz")

    local_ds.diff(a)
    dataset_changes_from_a = {"info_updated": True, "deleted": ["abc"]}
    target = get_diff_helper(
        dataset_changes_from_a, {}, {}, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == ({}, {})

    # cover DatasetDiff.frombuffer
    ds = hub.load(local_ds.path)
    ds.diff(a)
    captured = capsys.readouterr()
    assert captured.out == target
    diff = ds.diff(a, as_dict=True)["tensor"]
    assert diff == ({}, {})


def test_clear_diff(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append([1, 2, 3])
        local_ds.abc.clear()
        local_ds.abc.append([4, 5, 6])
        local_ds.abc.append([1, 2, 3])

    local_ds.diff()
    tensor_changes = {"abc": tensor_diff_helper([0, 2], created=True)}
    target = get_diff_helper({}, {}, tensor_changes, None)
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == tensor_changes

    a = local_ds.commit()

    with local_ds:
        local_ds.abc.append([3, 4, 5])
        local_ds.create_tensor("xyz")
        local_ds.abc.clear()
        local_ds.abc.append([1, 0, 0])
        local_ds.xyz.append([0, 1, 0])

    local_ds.diff(a)
    tensor_changes_from_a = {
        "abc": tensor_diff_helper([0, 1], cleared=True),
        "xyz": tensor_diff_helper([0, 1], created=True),
    }
    target = get_diff_helper(
        {}, {}, tensor_changes_from_a, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == (tensor_changes_from_a, {})

    b = local_ds.commit()

    with local_ds:
        local_ds.xyz.append([0, 0, 1])
        local_ds.xyz.clear()
        local_ds.xyz.append([1, 2, 3])
        local_ds.abc.append([3, 4, 2])
    c = local_ds.commit()

    local_ds.diff(c, a)
    tensor_changes_from_a = {
        "abc": tensor_diff_helper([0, 2], cleared=True),
        "xyz": tensor_diff_helper([0, 1], created=True),
    }
    target = get_diff_helper(
        {}, {}, tensor_changes_from_a, {}, local_ds.version_state, c, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == (tensor_changes_from_a, {})


def test_delete_diff(local_ds, capsys):
    local_ds.create_tensor("x/y/z")
    local_ds["x/y/z"].append([4, 5, 6])
    a = local_ds.commit()
    local_ds.create_tensor("a/b/c")
    b = local_ds.commit()
    local_ds["a/b/c"].append([1, 2, 3])
    c = local_ds.commit()
    local_ds.delete_tensor("a/b/c")

    local_ds.diff(a)
    target = get_diff_helper({}, {}, {}, {}, local_ds.version_state, a)
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == {}, {}

    d = local_ds.commit()
    local_ds["x/y/z"][0] = [1, 3, 4]
    e = local_ds.commit()
    local_ds.create_tensor("a/b/c")
    local_ds.delete_tensor("x/y/z")

    local_ds.diff(c)
    ds_changes_f_from_c = {"deleted": ["x/y/z", "a/b/c"]}
    tensor_changes_f_from_c = {"a/b/c": tensor_diff_helper(created=True)}
    target = get_diff_helper(
        ds_changes_f_from_c, {}, tensor_changes_f_from_c, {}, local_ds.version_state, c
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == tensor_changes_f_from_c, {}

    local_ds.diff(a)
    ds_changes_from_a = {"deleted": ["x/y/z"]}
    target = get_diff_helper(
        ds_changes_from_a, {}, tensor_changes_f_from_c, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == tensor_changes_f_from_c, {}


def test_rename_diff_single(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append([1, 2, 3])
        local_ds.rename_tensor("abc", "xyz")
        local_ds.xyz.append([2, 3, 4])
        local_ds.rename_tensor("xyz", "efg")
        local_ds.create_tensor("red")

    local_ds.diff()
    tensor_changes = {
        "efg": tensor_diff_helper([0, 2], created=True),
        "red": tensor_diff_helper(created=True),
    }
    target = get_diff_helper({}, {}, tensor_changes, None)
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == tensor_changes

    a = local_ds.commit()
    with local_ds:
        local_ds.rename_tensor("red", "blue")
        local_ds.efg.append([3, 4, 5])
        local_ds.rename_tensor("efg", "bcd")
        local_ds.bcd[1] = [2, 5, 4]
        local_ds.rename_tensor("bcd", "red")
        local_ds.red.append([1, 3, 4])
        local_ds.blue.append([2, 3, 4])
        local_ds.rename_tensor("blue", "efg")
    local_ds.diff(a)
    dataset_changes = {"renamed": OrderedDict({"red": "efg", "efg": "red"})}
    tensor_changes = {
        "red": tensor_diff_helper([2, 4], {1}),
        "efg": tensor_diff_helper([0, 1]),
    }
    target = get_diff_helper(
        dataset_changes, {}, tensor_changes, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == (tensor_changes, {})


def test_rename_diff_linear(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append([1, 2, 3])
        local_ds.create_tensor("xyz")

    a = local_ds.commit()
    with local_ds:
        local_ds.create_tensor("red")
        local_ds.xyz.append([3, 4, 5])
        local_ds.rename_tensor("xyz", "efg")
        local_ds.rename_tensor("abc", "xyz")
        local_ds.xyz[0] = [2, 3, 4]

    b = local_ds.commit()
    with local_ds:
        local_ds.rename_tensor("red", "blue")
        local_ds.xyz.append([5, 6, 7])
        local_ds.xyz.info["hello"] = "world"
        local_ds.rename_tensor("efg", "abc")
        local_ds.abc.append([6, 7, 8])

    local_ds.diff(a)
    ds_changes_from_a = {"renamed": OrderedDict({"xyz": "abc", "abc": "xyz"})}
    tensor_changes_from_a = {
        "xyz": tensor_diff_helper([1, 2], {0}, info_updated=True),
        "abc": tensor_diff_helper([0, 2]),
        "blue": tensor_diff_helper(created=True),
    }
    target = get_diff_helper(
        ds_changes_from_a, {}, tensor_changes_from_a, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == (tensor_changes_from_a, {})

    c = local_ds.commit()
    with local_ds:
        local_ds.rename_tensor("abc", "bcd")
        local_ds.rename_tensor("xyz", "abc")
        local_ds.delete_tensor("blue")

    local_ds.diff(a)
    ds_changes_from_a = {"renamed": OrderedDict({"xyz": "bcd"})}
    tensor_changes_from_a = {
        "abc": tensor_diff_helper([1, 2], {0}, info_updated=True),
        "bcd": tensor_diff_helper([0, 2]),
    }
    target = get_diff_helper(
        ds_changes_from_a, {}, tensor_changes_from_a, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == (tensor_changes_from_a, {})

    d = local_ds.commit()
    with local_ds:
        local_ds.delete_tensor("bcd")
        local_ds.rename_tensor("abc", "bcd")
        local_ds.bcd.append([4, 5, 6])

    local_ds.diff(b)
    ds_changes_from_b = {
        "renamed": OrderedDict({"xyz": "bcd"}),
        "deleted": ["efg", "red"],
    }
    tensor_changes_from_b = {
        "bcd": tensor_diff_helper([1, 3], info_updated=True),
    }
    target = get_diff_helper(
        ds_changes_from_b, {}, tensor_changes_from_b, {}, local_ds.version_state, b
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(b, as_dict=True)["tensor"]
    assert diff == (tensor_changes_from_b, {})

    e = local_ds.commit()
    with local_ds:
        local_ds.rename_tensor("bcd", "abc")

    local_ds.diff(a)
    ds_changes_from_a = {"deleted": ["xyz"]}
    tensor_changes_from_a = {"abc": tensor_diff_helper([1, 3], {0}, info_updated=True)}
    target = get_diff_helper(
        ds_changes_from_a, {}, tensor_changes_from_a, {}, local_ds.version_state, a
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert diff == (tensor_changes_from_a, {})


def test_rename_diff_branch(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("abc")
        local_ds.abc.append([1, 2, 3])

    a = local_ds.commit()
    local_ds.checkout("alt", create=True)

    with local_ds:
        local_ds.rename_tensor("abc", "xyz")
        local_ds.xyz.append([4, 5, 6])

    b = local_ds.commit()
    local_ds.checkout("main")

    with local_ds:
        local_ds.abc.append([2, 3, 4])
        local_ds.create_tensor("red")

    c = local_ds.commit()
    local_ds.checkout("alt2", create=True)

    with local_ds:
        local_ds.rename_tensor("abc", "efg")
        local_ds.efg.append([5, 6, 7])
        local_ds.efg.info["hello"] = "world"
        local_ds.rename_tensor("red", "blue")

    d = local_ds.commit()

    local_ds.delete_tensor("blue")

    e = local_ds.commit()

    local_ds.diff(b, e)

    ds_changes_b_from_a = {"renamed": OrderedDict({"abc": "xyz"})}
    tensor_changes_b_from_a = {"xyz": tensor_diff_helper([1, 2])}

    ds_changes_e_from_a = {"renamed": OrderedDict({"abc": "efg"})}
    tensor_changes_e_from_a = {"efg": tensor_diff_helper([1, 3], info_updated=True)}

    target = get_diff_helper(
        ds_changes_b_from_a,
        ds_changes_e_from_a,
        tensor_changes_b_from_a,
        tensor_changes_e_from_a,
        local_ds.version_state,
        b,
        e,
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(b, e, as_dict=True)["tensor"]
    assert diff == (tensor_changes_b_from_a, tensor_changes_e_from_a)


def test_rename_group(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("g1/g2/g3/t1")
        local_ds.create_tensor("g1/g2/t2")
        local_ds.commit()
        local_ds.rename_group("g1/g2", "g1/g4")

    diff = local_ds.diff()
    ds_changes = {
        "deleted": [],
        "info_updated": False,
        "renamed": OrderedDict({"g1/g2/g3/t1": "g1/g4/g3/t1", "g1/g2/t2": "g1/g4/t2"}),
    }
    target = get_diff_helper(ds_changes, {}, {}, None)
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["dataset"]
    assert diff == ds_changes


def test_diff_linear(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("xyz")
        local_ds.xyz.extend([1, 2, 3])
        local_ds.create_tensor("pqr")
        local_ds.pqr.extend([4, 5, 6])
    a = local_ds.commit()
    with local_ds:
        local_ds.xyz[0] = 10
        local_ds.xyz.info["hello"] = "world"
        local_ds.pqr[2] = 20
        local_ds.create_tensor("abc")
        local_ds.abc.extend([1, 2, 3])

    local_ds.diff()
    changes_b_from_a = {
        "xyz": {
            "data_added": [3, 3],
            "data_updated": {0},
            "created": False,
            "cleared": False,
            "info_updated": True,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "pqr": {
            "data_added": [3, 3],
            "data_updated": {2},
            "created": False,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "abc": {
            "data_added": [0, 3],
            "data_updated": set(),
            "created": True,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }
    message0 = NO_COMMIT_PASSED_DIFF
    message1 = "Diff in HEAD relative to the previous commit:\n"
    # ds_changes_1, ds_changes_2, tensor_changes_1, tensor_changes_2, msg_0, msg_1, msg_2
    target = (
        get_all_changes_string({}, {}, changes_b_from_a, None, message0, message1, None)
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == changes_b_from_a

    b = local_ds.commit()
    local_ds.diff()
    changes_empty = {}
    target = (
        get_all_changes_string({}, {}, changes_empty, None, message0, message1, None)
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == changes_empty

    local_ds.diff(a)
    lca_id = get_lca_id_helper(local_ds.version_state, a)
    message0 = ONE_COMMIT_PASSED_DIFF % lca_id
    message1 = "Diff in HEAD:\n"
    message2 = f"Diff in {a} (target id):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_b_from_a, changes_empty, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_b_from_a
    assert diff[1] == changes_empty

    local_ds.diff(b)
    lca_id = get_lca_id_helper(local_ds.version_state, b)
    message0 = ONE_COMMIT_PASSED_DIFF % lca_id
    message2 = f"Diff in {b} (target id):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_empty, changes_empty, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(b, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_empty
    assert diff[1] == changes_empty

    local_ds.diff(a, b)
    lca_id = get_lca_id_helper(local_ds.version_state, a, b)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {a} (target id 1):\n"
    message2 = f"Diff in {b} (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_empty, changes_b_from_a, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, b, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_empty
    assert diff[1] == changes_b_from_a

    local_ds.diff(b, a)
    lca_id = get_lca_id_helper(local_ds.version_state, b, a)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {b} (target id 1):\n"
    message2 = f"Diff in {a} (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_b_from_a, changes_empty, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(b, a, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_b_from_a
    assert diff[1] == changes_empty

    local_ds.checkout(b)
    local_ds.diff()
    message0 = NO_COMMIT_PASSED_DIFF
    message1 = f"Diff in {b} (current commit) relative to the previous commit:\n"
    target = (
        get_all_changes_string({}, {}, changes_b_from_a, None, message0, message1, None)
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == changes_b_from_a

    local_ds.diff(a)
    lca_id = get_lca_id_helper(local_ds.version_state, a)
    message0 = ONE_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {b} (current commit):\n"
    message2 = f"Diff in {a} (target id):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_b_from_a, changes_empty, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)


def test_diff_branch(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("xyz")
        local_ds.xyz.extend([1, 2, 3])

    a = local_ds.commit()
    local_ds.checkout("alt", create=True)
    with local_ds:
        local_ds.xyz.extend([4, 5, 6])
        local_ds.create_tensor("pqr")
        local_ds.pqr.extend([7, 8, 9])
        local_ds.xyz[2] = 6
        local_ds.xyz[3] = 8
        local_ds.pqr[1] = 8

    b = local_ds.commit()
    local_ds.checkout("main")
    with local_ds:
        local_ds.xyz.extend([0, 0])
        local_ds.xyz[2] = 10
        local_ds.xyz[3] = 11
        local_ds.xyz[0] = 11

    local_ds.diff()
    changes_b_from_branch_off = {
        "xyz": {
            "data_added": [3, 6],
            "data_updated": {2},
            "created": False,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "pqr": {
            "data_added": [0, 3],
            "data_updated": set(),
            "created": True,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }
    changes_main_from_branch_off = {
        "xyz": {
            "data_added": [3, 5],
            "data_updated": {0, 2},
            "created": False,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }
    message0 = NO_COMMIT_PASSED_DIFF
    message1 = "Diff in HEAD relative to the previous commit:\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_main_from_branch_off, None, message0, message1, None
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == changes_main_from_branch_off

    c = local_ds.commit()

    local_ds.diff()
    empty_changes = {}
    target = (
        get_all_changes_string({}, {}, empty_changes, None, message0, message1, None)
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(as_dict=True)["tensor"]
    assert diff == empty_changes

    local_ds.diff(a)
    lca_id = get_lca_id_helper(local_ds.version_state, a)
    message0 = ONE_COMMIT_PASSED_DIFF % lca_id
    message1 = "Diff in HEAD:\n"
    message2 = f"Diff in {a} (target id):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            changes_main_from_branch_off,
            empty_changes,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_main_from_branch_off
    assert diff[1] == empty_changes

    local_ds.diff(b)
    lca_id = get_lca_id_helper(local_ds.version_state, b)
    message0 = ONE_COMMIT_PASSED_DIFF % lca_id
    message2 = f"Diff in {b} (target id):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            changes_main_from_branch_off,
            changes_b_from_branch_off,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(b, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_main_from_branch_off
    assert diff[1] == changes_b_from_branch_off

    local_ds.diff(c)
    lca_id = get_lca_id_helper(local_ds.version_state, c)
    message0 = ONE_COMMIT_PASSED_DIFF % lca_id
    message2 = f"Diff in {c} (target id):\n"
    target = (
        get_all_changes_string(
            {}, {}, empty_changes, empty_changes, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(c, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == empty_changes
    assert diff[1] == empty_changes

    local_ds.diff(a, b)
    lca_id = get_lca_id_helper(local_ds.version_state, a)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {a} (target id 1):\n"
    message2 = f"Diff in {b} (target id 2):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            empty_changes,
            changes_b_from_branch_off,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, b, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == empty_changes
    assert diff[1] == changes_b_from_branch_off

    local_ds.diff(b, a)
    lca_id = get_lca_id_helper(local_ds.version_state, a, b)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {b} (target id 1):\n"
    message2 = f"Diff in {a} (target id 2):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            changes_b_from_branch_off,
            empty_changes,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(b, a, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_b_from_branch_off
    assert diff[1] == empty_changes

    local_ds.diff(b, c)
    lca_id = get_lca_id_helper(local_ds.version_state, b, c)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {b} (target id 1):\n"
    message2 = f"Diff in {c} (target id 2):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            changes_b_from_branch_off,
            changes_main_from_branch_off,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target

    local_ds.diff(c, b)
    lca_id = get_lca_id_helper(local_ds.version_state, c, b)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {c} (target id 1):\n"
    message2 = f"Diff in {b} (target id 2):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            changes_main_from_branch_off,
            changes_b_from_branch_off,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(c, b, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_main_from_branch_off
    assert diff[1] == changes_b_from_branch_off

    local_ds.diff(c, a)
    lca_id = get_lca_id_helper(local_ds.version_state, c, a)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {c} (target id 1):\n"
    message2 = f"Diff in {a} (target id 2):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            changes_main_from_branch_off,
            empty_changes,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(c, a, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_main_from_branch_off
    assert diff[1] == empty_changes

    local_ds.diff(a, c)
    lca_id = get_lca_id_helper(local_ds.version_state, a, c)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {a} (target id 1):\n"
    message2 = f"Diff in {c} (target id 2):\n"
    target = (
        get_all_changes_string(
            {},
            {},
            empty_changes,
            changes_main_from_branch_off,
            message0,
            message1,
            message2,
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(a, c, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == empty_changes
    assert diff[1] == changes_main_from_branch_off


def test_complex_diff(local_ds, capsys):
    with local_ds:
        local_ds.create_tensor("xyz")
        local_ds.xyz.extend([1, 2, 3])
    a = local_ds.commit()
    b = local_ds.checkout("alt", create=True)
    with local_ds:
        local_ds.xyz.extend([4, 5, 6])
    local_ds.commit()
    c = local_ds.pending_commit_id
    with local_ds:
        local_ds.xyz[4] = 7
        local_ds.xyz[0] = 0
    local_ds.checkout("main")
    d = local_ds.pending_commit_id
    with local_ds:
        local_ds.xyz[1] = 10
        local_ds.create_tensor("pqr")
    local_ds.commit()
    f = local_ds.checkout("another", create=True)
    with local_ds:
        local_ds.create_tensor("tuv")
        local_ds.tuv.extend([1, 2, 3])
        local_ds.pqr.append(5)
    local_ds.commit()
    g = local_ds.pending_commit_id
    e = local_ds.checkout("main")

    # x is LCA of a and g
    changes_c_from_x = {
        "xyz": {
            "data_added": [3, 6],
            "data_updated": {0},
            "created": False,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }
    changes_g_from_x = {
        "pqr": {
            "data_added": [0, 1],
            "data_updated": set(),
            "created": True,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "tuv": {
            "data_added": [0, 3],
            "data_updated": set(),
            "created": True,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "xyz": {
            "data_added": [3, 3],
            "data_updated": {1},
            "created": False,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }
    empty_changes = {}

    local_ds.diff(c, g)
    lca_id = get_lca_id_helper(local_ds.version_state, c, g)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {c} (target id 1):\n"
    message2 = f"Diff in {g} (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_c_from_x, changes_g_from_x, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(c, g, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_c_from_x
    assert diff[1] == changes_g_from_x

    local_ds.diff(e, d)
    lca_id = get_lca_id_helper(local_ds.version_state, e, d)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {e} (target id 1):\n"
    message2 = f"Diff in {d} (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, empty_changes, empty_changes, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(e, d, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == empty_changes
    assert diff[1] == empty_changes

    local_ds.diff(e, e)
    lca_id = get_lca_id_helper(local_ds.version_state, e, e)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {e} (target id 1):\n"
    message2 = f"Diff in {e} (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, empty_changes, empty_changes, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(e, e, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == empty_changes
    assert diff[1] == empty_changes

    changes_main_from_x = {
        "xyz": {
            "data_added": [3, 3],
            "data_updated": {1},
            "created": False,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "pqr": {
            "data_added": [0, 0],
            "data_updated": set(),
            "created": True,
            "cleared": False,
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }

    local_ds.diff(c, "main")
    lca_id = get_lca_id_helper(local_ds.version_state, c, "main")
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = f"Diff in {c} (target id 1):\n"
    message2 = "Diff in main (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_c_from_x, changes_main_from_x, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff(c, "main", as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_c_from_x
    assert diff[1] == changes_main_from_x

    local_ds.diff("main", c)
    lca_id = get_lca_id_helper(local_ds.version_state, "main", c)
    message0 = TWO_COMMIT_PASSED_DIFF % lca_id
    message1 = "Diff in main (target id 1):\n"
    message2 = f"Diff in {c} (target id 2):\n"
    target = (
        get_all_changes_string(
            {}, {}, changes_main_from_x, changes_c_from_x, message0, message1, message2
        )
        + "\n"
    )
    captured = capsys.readouterr()
    assert captured.out == target
    diff = local_ds.diff("main", c, as_dict=True)["tensor"]
    assert isinstance(diff, tuple)
    assert diff[0] == changes_main_from_x
    assert diff[1] == changes_c_from_x


def test_diff_not_exists(local_ds):
    with pytest.raises(KeyError):
        local_ds.diff("12345", "5678")


def test_branches(local_ds_generator):
    local_ds = local_ds_generator()
    assert local_ds.branches == ["main"]
    local_ds.checkout("alt", create=True)
    assert local_ds.branches == ["main", "alt"]
    local_ds.checkout("main")
    assert local_ds.branches == ["main", "alt"]

    local_ds = local_ds_generator()
    assert local_ds.branches == ["main", "alt"]
    local_ds.checkout("alt")
    assert local_ds.branches == ["main", "alt"]
    local_ds.checkout("other", create=True)
    assert local_ds.branches == ["main", "alt", "other"]


def test_commits(local_ds):
    commits = local_ds.commits
    assert len(commits) == 0
    local_ds.commit()
    commits = local_ds.commits
    assert len(commits) == 1
    commit_details_helper(commits, local_ds)
    local_ds.checkout("alt", create=True)
    commits = local_ds.commits
    assert len(commits) == 1
    commit_details_helper(commits, local_ds)
    local_ds.checkout("main")
    commits = local_ds.commits
    assert len(commits) == 1
    commit_details_helper(commits, local_ds)
    local_ds.create_tensor("xyz")
    local_ds.checkout("other", create=True)
    commits = local_ds.commits
    assert len(commits) == 2
    commit_details_helper(commits, local_ds)
    local_ds.commit(allow_empty=True)
    commits = local_ds.commits
    assert len(commits) == 3
    commit_details_helper(commits, local_ds)
    local_ds.checkout("main")
    commits = local_ds.commits
    assert len(commits) == 2
    commit_details_helper(commits, local_ds)
    local_ds.commit(allow_empty=True)
    commits = local_ds.commits
    assert len(commits) == 3
    commit_details_helper(commits, local_ds)


def test_clear(local_ds):
    local_ds.create_tensor("abc")
    local_ds.abc.append([1, 2, 3])
    a = local_ds.commit("first")
    local_ds.abc.clear()
    b = local_ds.commit("second")
    local_ds.abc.append([4, 5, 6, 7])
    c = local_ds.commit("third")
    local_ds.abc.clear()

    assert len(local_ds.abc.numpy()) == 0

    local_ds.checkout(a)

    np.testing.assert_array_equal(local_ds.abc.numpy(), np.array([[1, 2, 3]]))

    local_ds.checkout(b)

    assert len(local_ds.abc.numpy()) == 0

    local_ds.checkout(c)

    np.testing.assert_array_equal(local_ds.abc.numpy(), np.array([[4, 5, 6, 7]]))


def test_custom_commit_hash(local_ds):
    commits = local_ds.commits
    assert len(commits) == 0
    local_ds._commit(hash="abcd")
    assert local_ds.version_state["commit_id"] == "abcd"
    with pytest.raises(CommitError):
        local_ds._commit(hash="abcd")
    with pytest.raises(CommitError):
        local_ds._checkout("xyz", create=True, hash="abcd")
    local_ds._checkout("xyz", create=True, hash="efgh")
    assert local_ds.version_state["commit_id"] == "efgh"
    assert set(local_ds.version_state["branch_commit_map"].keys()) == set(
        ("main", "xyz")
    )
    assert local_ds.version_state["branch_commit_map"]["xyz"] == "efgh"


def test_read_only_checkout(local_ds):
    with local_ds:
        local_ds.create_tensor("x")
        local_ds.x.append([1, 2, 3])
        local_ds.checkout("branch", create=True)
        local_ds.checkout("main")
    assert local_ds.storage.autoflush == True
    local_ds.read_only = True
    local_ds.checkout("main")


def test_modified_samples(memory_ds):
    with memory_ds:
        memory_ds.create_tensor("image")
        memory_ds.image.extend(np.array(list(range(5))))

        img, indexes = memory_ds.image.modified_samples(return_indexes=True)
        assert indexes == list(range(5))
        assert len(img) == 5
        for i in range(5):
            np.testing.assert_array_equal(img[i].numpy(), i)
        first_commit = memory_ds.commit()

        img, indexes = memory_ds.image.modified_samples(return_indexes=True)
        assert indexes == []
        assert len(img) == 0

        memory_ds.image.extend(np.array(list(range(5, 8))))
        img, indexes = memory_ds.image.modified_samples(return_indexes=True)
        assert indexes == list(range(5, 8))
        assert len(img) == 3
        for i in range(3):
            np.testing.assert_array_equal(img[i].numpy(), i + 5)

        memory_ds.image[2] = -1
        img, indexes = memory_ds.image.modified_samples(return_indexes=True)
        assert indexes == [2, 5, 6, 7]
        assert len(img) == 4
        np.testing.assert_array_equal(img[0].numpy(), -1)
        for i in range(3):
            np.testing.assert_array_equal(img[i + 1].numpy(), i + 5)

        memory_ds.image[4] = 8
        img, indexes = memory_ds.image.modified_samples(return_indexes=True)
        assert indexes == [2, 4, 5, 6, 7]
        assert len(img) == 5
        np.testing.assert_array_equal(img[0].numpy(), -1)
        np.testing.assert_array_equal(img[1].numpy(), 8)
        for i in range(3):
            np.testing.assert_array_equal(img[i + 2].numpy(), i + 5)

        second_commit = memory_ds.commit()
        img = memory_ds.image.modified_samples()
        assert len(img) == 0

        img, indexes = memory_ds.image.modified_samples(
            first_commit, return_indexes=True
        )
        assert indexes == [2, 4, 5, 6, 7]
        assert len(img) == 5
        np.testing.assert_array_equal(img[0].numpy(), -1)
        np.testing.assert_array_equal(img[1].numpy(), 8)
        for i in range(3):
            np.testing.assert_array_equal(img[i + 2].numpy(), i + 5)

        memory_ds.checkout(first_commit)
        memory_ds.checkout("alt", create=True)
        alt_commit = memory_ds.commit(allow_empty=True)

        memory_ds.checkout("main")

        with pytest.raises(TensorModifiedError):
            memory_ds.image.modified_samples(alt_commit)


def test_reset(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc")
        ds.create_tensor("xyz")
        for i in range(10):
            ds.abc.append(i)
            ds.xyz.append(np.ones((100, 100, 3)) * i)
        ds.info.hello = "world"
        ds.abc.info.hello = "world"
        assert list(local_ds.tensors) == ["abc", "xyz"]
        assert local_ds.info.hello == "world"
        for i in range(10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), i)
            np.testing.assert_array_equal(ds.xyz[i].numpy(), np.ones((100, 100, 3)) * i)

        ds.reset()
        assert not list(local_ds.tensors)
        assert "hello" not in local_ds.info

        ds.create_tensor("abc")
        for i in range(10):
            ds.abc.append(i)
        ds.info.hello = "world"
        ds.abc.info.hello = "world"
        assert list(local_ds.tensors) == ["abc"]
        assert local_ds.info.hello == "world"
        assert local_ds.abc.info.hello == "world"
        assert len(ds) == 10

        for i in range(10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), i)

        ds.commit()
        ds.reset()

        assert list(local_ds.tensors) == ["abc"]
        assert local_ds.info.hello == "world"
        assert local_ds.abc.info.hello == "world"
        for i in range(10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), i)

        for i in range(10):
            ds.abc[i] = 0
        for i in range(10, 15):
            ds.abc.append(i)

        ds.info.hello = "new world"
        ds.abc.info.hello1 = "world1"

        assert ds.info.hello == "new world"
        assert local_ds.abc.info.hello == "world"
        assert ds.abc.info.hello1 == "world1"
        assert len(ds) == 15

        for i in range(10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), 0)

        for i in range(10, 15):
            np.testing.assert_array_equal(ds.abc[i].numpy(), i)

        ds.reset()

        assert list(local_ds.tensors) == ["abc"]
        assert local_ds.info.hello == "world"
        assert local_ds.abc.info.hello == "world"
        assert "hello1" not in ds.abc.info
        assert len(ds) == 10

        for i in range(10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), i)


def test_reset_create_delete_tensors(local_ds):
    with local_ds as ds:
        local_ds.create_tensor("one")
        local_ds.create_tensor("two")
        assert set(ds.tensors.keys()) == {"one", "two"}
        ds.commit()
        assert set(ds.tensors.keys()) == {"one", "two"}
        ds.create_tensor("three")
        ds.delete_tensor("two")
        assert set(ds.tensors.keys()) == {"one", "three"}
        ds.reset()
        assert set(ds.tensors.keys()) == {"one", "two"}


@pytest.mark.parametrize(
    "ds_generator",
    [
        "local_ds_generator",
        "s3_ds_generator",
        "gcs_ds_generator",
        "hub_cloud_ds_generator",
    ],
    indirect=True,
)
def test_reset_bug(ds_generator):
    ds = ds_generator()
    ds.create_tensor("abc")
    ds.abc.append([1, 2, 3])
    assert len(ds.abc) == 1
    a = ds.commit()

    ds = ds_generator()
    ds.abc.append([3, 4, 5])
    assert len(ds.abc) == 2
    ds.reset()
    assert len(ds.abc) == 1


def test_reset_delete_group(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc/x")
        ds.commit()
        ds.delete_group("abc")
        assert ds.has_head_changes



import posixpath
from hub.constants import (
    CHUNKS_FOLDER,
    DATASET_DIFF_FILENAME,
    DATASET_INFO_FILENAME,
    DATASET_LOCK_FILENAME,
    ENCODED_CREDS_FOLDER,
    LINKED_CREDS_FILENAME,
    UNSHARDED_ENCODER_FILENAME,
    ENCODED_CHUNK_NAMES_FOLDER,
    ENCODED_SEQUENCE_NAMES_FOLDER,
    ENCODED_TILE_NAMES_FOLDER,
    FIRST_COMMIT_ID,
    DATASET_META_FILENAME,
    TENSOR_INFO_FILENAME,
    TENSOR_META_FILENAME,
    TENSOR_COMMIT_CHUNK_SET_FILENAME,
    TENSOR_COMMIT_DIFF_FILENAME,
    VERSION_CONTROL_INFO_FILENAME,
    VERSION_CONTROL_INFO_FILENAME_OLD,
    VERSION_CONTROL_INFO_LOCK_FILENAME,
    QUERIES_FILENAME,
    QUERIES_LOCK_FILENAME,
)
from hub.util.exceptions import S3GetError


def get_chunk_key(key: str, chunk_name: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join((key, CHUNKS_FOLDER, f"{chunk_name}"))

    return "/".join(("versions", commit_id, key, CHUNKS_FOLDER, f"{chunk_name}"))


def get_dataset_meta_key(commit_id: str) -> str:
    # dataset meta is always relative to the `StorageProvider`'s root
    if commit_id == FIRST_COMMIT_ID:
        return DATASET_META_FILENAME

    return "/".join(("versions", commit_id, DATASET_META_FILENAME))


def get_dataset_info_key(commit_id: str) -> str:
    # dataset info is always relative to the `StorageProvider`'s root
    if commit_id == FIRST_COMMIT_ID:
        return DATASET_INFO_FILENAME
    return "/".join(("versions", commit_id, DATASET_INFO_FILENAME))


def get_dataset_diff_key(commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return DATASET_DIFF_FILENAME
    return "/".join(("versions", commit_id, DATASET_DIFF_FILENAME))


def get_dataset_linked_creds_key() -> str:
    return LINKED_CREDS_FILENAME


def get_dataset_linked_creds_lock_key() -> str:
    return VERSION_CONTROL_INFO_LOCK_FILENAME


def get_version_control_info_key() -> str:
    return VERSION_CONTROL_INFO_FILENAME


def get_version_control_info_key_old() -> str:
    return VERSION_CONTROL_INFO_FILENAME_OLD


def get_version_control_info_lock_key() -> str:
    return VERSION_CONTROL_INFO_LOCK_FILENAME


def get_dataset_lock_key() -> str:
    return DATASET_LOCK_FILENAME


def get_tensor_meta_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join((key, TENSOR_META_FILENAME))
    return "/".join(("versions", commit_id, key, TENSOR_META_FILENAME))


def get_tensor_tile_encoder_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join((key, ENCODED_TILE_NAMES_FOLDER, UNSHARDED_ENCODER_FILENAME))
    return "/".join(
        (
            "versions",
            commit_id,
            key,
            ENCODED_TILE_NAMES_FOLDER,
            UNSHARDED_ENCODER_FILENAME,
        )
    )


def get_creds_encoder_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join((key, ENCODED_CREDS_FOLDER, UNSHARDED_ENCODER_FILENAME))
    return "/".join(
        (
            "versions",
            commit_id,
            key,
            ENCODED_CREDS_FOLDER,
            UNSHARDED_ENCODER_FILENAME,
        )
    )


def get_tensor_info_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join([key, TENSOR_INFO_FILENAME])
    return "/".join(("versions", commit_id, key, TENSOR_INFO_FILENAME))


def get_tensor_commit_chunk_set_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join((key, TENSOR_COMMIT_CHUNK_SET_FILENAME))
    return "/".join(("versions", commit_id, key, TENSOR_COMMIT_CHUNK_SET_FILENAME))


def get_tensor_commit_diff_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join((key, "commit_diff"))
    return "/".join(("versions", commit_id, key, TENSOR_COMMIT_DIFF_FILENAME))


def get_chunk_id_encoder_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join(
            (
                key,
                ENCODED_CHUNK_NAMES_FOLDER,
                UNSHARDED_ENCODER_FILENAME,
            )
        )
    return "/".join(
        (
            "versions",
            commit_id,
            key,
            ENCODED_CHUNK_NAMES_FOLDER,
            UNSHARDED_ENCODER_FILENAME,
        )
    )


def get_sequence_encoder_key(key: str, commit_id: str) -> str:
    if commit_id == FIRST_COMMIT_ID:
        return "/".join(
            (
                key,
                ENCODED_SEQUENCE_NAMES_FOLDER,
                UNSHARDED_ENCODER_FILENAME,
            )
        )
    return "/".join(
        (
            "versions",
            commit_id,
            key,
            ENCODED_SEQUENCE_NAMES_FOLDER,
            UNSHARDED_ENCODER_FILENAME,
        )
    )


def dataset_exists(storage) -> bool:
    try:
        storage[get_dataset_meta_key(FIRST_COMMIT_ID)]
        return True
    except (KeyError, S3GetError):
        return False


def tensor_exists(key: str, storage, commit_id: str) -> bool:
    try:
        storage[get_tensor_meta_key(key, commit_id)]
        return True
    except KeyError:
        return False


def get_queries_key() -> str:
    return QUERIES_FILENAME


def get_queries_lock_key() -> str:
    return QUERIES_LOCK_FILENAME


def filter_name(name: str, group_index: str = "") -> str:
    """Filters tensor name and returns full name of the tensor"""
    name = name.strip("/")

    while "//" in name:
        name = name.replace("//", "/")

    name = posixpath.join(group_index, name)
    return name


def get_sample_info_tensor_key(key: str):
    group, key = posixpath.split(key)
    return posixpath.join(group, f"_{key}_info")


def get_sample_id_tensor_key(key: str):
    group, key = posixpath.split(key)
    return posixpath.join(group, f"_{key}_id")


def get_sample_shape_tensor_key(key: str):
    group, key = posixpath.split(key)
    return posixpath.join(group, f"_{key}_shape")



import pathlib
from typing import Union, Set
from hub.core.dataset import Dataset
import posixpath
import hub
from tqdm import tqdm  # type: ignore


def _is_seq_convertible(seq):
    from datasets import Sequence

    if isinstance(seq, Sequence):
        feature = seq.feature
    else:
        feature = seq[0]
    if isinstance(feature, dict):
        return False
    if feature.dtype in (
        "string",
        "large_string",
    ):  # no support for sequences of strings.
        return False
    if isinstance(feature, Sequence):
        return _is_seq_convertible(feature)
    return True


def _create_tensor_from_feature(key, feature, src, ds):
    from datasets import Sequence, ClassLabel
    from datasets import Dataset as hfDataset

    curr = posixpath.split(key)[-1]
    if isinstance(feature, (dict, Sequence, list)):
        if isinstance(feature, dict):
            for x in feature:
                _create_tensor_from_feature(f"{key}/{x}", feature[x], src[curr], ds)
            return True
        elif isinstance(feature, (Sequence, list)):
            if isinstance(feature, Sequence):
                inner = feature.feature
            else:
                inner = feature[0]
            if isinstance(inner, dict):
                _create_tensor_from_feature(key, feature.feature, src, ds)
                return True
            elif _is_seq_convertible(feature):
                ds.create_tensor(key)
            else:
                return False
    elif feature.dtype in ("string", "large_string"):
        ds.create_tensor(key, htype="text")
    elif isinstance(feature, ClassLabel):
        ds.create_tensor(key, htype="class_label", class_names=feature.names)
    else:
        ds.create_tensor(key)

    if isinstance(src, (hfDataset, dict)):
        ds[key].extend(src[curr])
    else:
        values = []
        for i in range(len(src)):
            values.extend(src[i][curr])
        ds[key].extend(values)
    return True


def ingest_huggingface(
    src,
    dest,
    use_progressbar=True,
) -> Dataset:
    """Converts hugging face datasets to hub format.

    Args:
        src (hfDataset, DatasetDict): Hugging Face Dataset or DatasetDict to be converted. Data in different splits of a
            DatasetDict will be stored under respective tensor groups.
        dest (Dataset, str, pathlib.Path): Destination dataset or path to it.
        use_progressbar (bool): Defines if progress bar should be used to show conversion progress.

    Returns:
        Dataset: The destination Hub dataset.

    Note:
        - if DatasetDict looks like:

            >>> {
            ...    train: Dataset({
            ...        features: ['data']
            ...    }),
            ...    validation: Dataset({
            ...        features: ['data']
            ...    }),
            ...    test: Dataset({
            ...        features: ['data']
            ...    }),
            ... }

        it will be converted to a Hub :class:`Dataset` with tensors ``['train/data', 'validation/data', 'test/data']``.

        Features of the type ``Sequence(feature=Value(dtype='string'))`` are not supported. Columns of such type are skipped.

    """
    from datasets import DatasetDict

    if isinstance(dest, (str, pathlib.Path)):
        ds = hub.dataset(dest)
    else:
        ds = dest  # type: ignore

    if isinstance(src, DatasetDict):
        for split, src_ds in src.items():
            for key, feature in src_ds.features.items():
                _create_tensor_from_feature(f"{split}/{key}", feature, src_ds, ds)
    else:
        skipped_keys: Set[str] = set()
        features = tqdm(
            src.features.items(),
            desc=f"Converting...({len(skipped_keys)} skipped)",
            disable=not use_progressbar,
        )
        for key, feature in features:
            if not _create_tensor_from_feature(key, feature, src, ds):
                skipped_keys.add(key)
                features.set_description(f"Converting...({len(skipped_keys)} skipped)")
    return ds  # type: ignore



import hub
import torch
import numpy as np
from hub.util.iterable_ordered_dict import IterableOrderedDict


def raise_indra_installation_error(indra_installed: bool):
    if not indra_installed:
        raise ImportError(
            "This is an experimental feature that requires Hub deeplake package. To use it, you can run `pip install hub[deeplake]`."
        )


def collate_fn(batch):
    elem = batch[0]

    if isinstance(elem, IterableOrderedDict):
        return IterableOrderedDict(
            (key, collate_fn([d[key] for d in batch])) for key in elem.keys()
        )
    if isinstance(elem, np.ndarray) and elem.dtype.type is np.str_:
        batch = [it.item() for it in batch]

    return torch.utils.data._utils.collate.default_collate(batch)



from PIL import Image  # type: ignore
from PIL.ExifTags import TAGS  # type: ignore
from hub.util.exif import getexif
import numpy as np
import pytest
import os
import sys
import hub
import json


def get_exif_helper(path):
    img = Image.open(path)
    return json.loads(json.dumps(getexif(img)))


def test_image_samples(local_ds_generator, compressed_image_paths):
    ds = local_ds_generator()
    jpg = ds.create_tensor("jpg_images", htype="image", sample_compression="jpg")
    jpg_paths = compressed_image_paths["jpeg"]
    for jpg_path in jpg_paths:
        ds.jpg_images.append(hub.read(jpg_path))

    for i, (jpg_path, sample_info) in enumerate(zip(jpg_paths, jpg.sample_info)):
        img = Image.open(jpg_path)
        sample_info2 = jpg[i].sample_info
        assert sample_info == sample_info2
        assert sample_info["exif"] == get_exif_helper(jpg_path)
        assert sample_info["shape"] == list(np.array(img).shape)
        assert sample_info["format"] == "jpeg"
        assert sample_info["filename"] == jpg_path

    ds = local_ds_generator()
    jpg = ds["jpg_images"]
    for i, (jpg_path, sample_info) in enumerate(zip(jpg_paths, jpg.sample_info)):
        img = Image.open(jpg_path)
        sample_info2 = jpg[i].sample_info
        assert sample_info == sample_info2
        assert sample_info["exif"] == get_exif_helper(jpg_path)
        assert sample_info["shape"] == list(np.array(img).shape)
        assert sample_info["format"] == "jpeg"
        assert sample_info["filename"] == jpg_path


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_video_samples(local_ds_generator, video_paths):
    import av  # type: ignore

    ds = local_ds_generator()
    mp4 = ds.create_tensor("mp4_videos", htype="video", sample_compression="mp4")
    mp4_paths = video_paths["mp4"]
    for mp4_path in mp4_paths:
        ds.mp4_videos.append(hub.read(mp4_path))

    for i, (mp4_path, sample_info) in enumerate(zip(mp4_paths, mp4.sample_info)):
        container = av.open(mp4_path)
        vstream = container.streams.video[0]

        sample_info2 = mp4[i].sample_info
        assert sample_info == sample_info2
        assert (
            sample_info["fps"]
            == vstream.guessed_rate.numerator / vstream.guessed_rate.denominator
        )
        assert sample_info["duration"] == vstream.duration or container.duration
        assert sample_info["format"] == "mp4"
        assert sample_info["filename"] == mp4_path

    ds = local_ds_generator()
    mp4 = ds["mp4_videos"]
    for i, (mp4_path, sample_info) in enumerate(zip(mp4_paths, mp4.sample_info)):
        container = av.open(mp4_path)
        vstream = container.streams.video[0]

        sample_info2 = mp4[i].sample_info
        assert sample_info == sample_info2
        assert (
            sample_info["fps"]
            == vstream.guessed_rate.numerator / vstream.guessed_rate.denominator
        )
        assert sample_info["duration"] == vstream.duration or container.duration
        assert sample_info["format"] == "mp4"
        assert sample_info["filename"] == mp4_path


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_audio_samples(local_ds_generator, audio_paths):
    import av  # type: ignore

    ds = local_ds_generator()
    mp3 = ds.create_tensor("mp3_audios", htype="audio", sample_compression="mp3")
    mp3_paths = [audio_paths["mp3"]]
    for mp3_path in mp3_paths:
        ds.mp3_audios.append(hub.read(mp3_path))

    for i, (mp3_path, sample_info) in enumerate(zip(mp3_paths, mp3.sample_info)):
        container = av.open(mp3_path)
        astream = container.streams.audio[0]

        sample_info2 = mp3[i].sample_info
        assert sample_info == sample_info2
        assert sample_info["nchannels"] == astream.channels
        assert sample_info["sample_rate"] == astream.sample_rate
        assert sample_info["duration"] == astream.duration or container.duration

    ds = local_ds_generator()
    mp3 = ds["mp3_audios"]
    for i, (mp3_path, sample_info) in enumerate(zip(mp3_paths, mp3.sample_info)):
        container = av.open(mp3_path)
        astream = container.streams.audio[0]

        sample_info2 = mp3[i].sample_info
        assert sample_info == sample_info2
        assert sample_info["nchannels"] == astream.channels
        assert sample_info["sample_rate"] == astream.sample_rate
        assert sample_info["duration"] == astream.duration or container.duration



import os
import sys
import pytest

import hub
from hub.core.dataset import Dataset
from hub.util.exceptions import DynamicTensorNumpyError

import numpy as np


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
@pytest.mark.parametrize("compression", hub.compression.VIDEO_COMPRESSIONS)
def test_video(local_ds, compression, video_paths):
    for i, path in enumerate(video_paths[compression]):
        tensor = local_ds.create_tensor(
            f"video_{i}", htype="video", sample_compression=compression
        )
        sample = hub.read(path)
        assert len(sample.shape) == 4
        if "dummy_data" in path:  # check shape only for internal test videos
            if compression == "mp4":
                assert sample.shape == (400, 360, 640, 3)
            elif compression == "mkv":
                assert sample.shape == (399, 360, 640, 3)
            elif compression == "avi":
                assert sample.shape == (901, 270, 480, 3)
        assert sample.shape[-1] == 3
        with local_ds:
            for _ in range(5):
                tensor.append(hub.read(path))  # type: ignore
            tensor.extend([hub.read(path) for _ in range(5)])  # type: ignore
        for i in range(10):
            assert tensor[i].numpy().shape == sample.shape  # type: ignore


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_video_slicing(local_ds: Dataset, video_paths):
    for path in video_paths["mp4"]:
        if "samplemp4_1MB" in path:
            dummy = np.zeros((132, 720, 1080, 3))

            local_ds.create_tensor("video", htype="video", sample_compression="mp4")
            local_ds.video.append(hub.read(path))
            local_ds.video[0][0:5].numpy().shape == dummy[0:5].shape
            local_ds.video[0][100:120].numpy().shape == dummy[100:120].shape
            local_ds.video[0][120].numpy().shape == dummy[120].shape
            local_ds.video[0][10:5:-2].numpy().shape == dummy[10:5:-2].shape
            local_ds.video[0][-3:-10:-1].numpy().shape == dummy[-3:-10:-1].shape
            local_ds.video[0][-25:100:-2].numpy().shape == dummy[-25:100:-2].shape
            local_ds.video[0][::-1].numpy().shape == dummy[::-1].shape
            local_ds.video[0][:5:-1].numpy().shape == dummy[:5:-1].shape
            local_ds.video[0][-1].numpy().shape == dummy[-1].shape
            return
    raise Exception  # test did not run


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
@pytest.mark.parametrize(
    ("vstream_path", "hub_token"),
    [
        ("gcs_vstream_path", "hub_cloud_dev_token"),
        ("s3_vstream_path", "hub_cloud_dev_token"),
        ("hub_cloud_vstream_path", "hub_cloud_dev_token"),
    ],
    indirect=True,
)
def test_video_streaming(vstream_path, hub_token):
    ds = hub.load(vstream_path, read_only=True, token=hub_token)

    # no streaming, downloads chunk
    assert ds.mp4_videos[0].shape == (400, 360, 640, 3)
    assert ds.mp4_videos[0].numpy().shape == (400, 360, 640, 3)
    assert ds.mp4_videos[1].numpy().shape == (120, 1080, 1920, 3)

    # streaming
    assert ds.large_video[0].shape == (21312, 546, 1280, 3)
    assert ds.large_video[0, 13500].numpy().shape == (546, 1280, 3)
    # will use cached url
    assert ds.large_video[0, 18000].numpy().shape == (546, 1280, 3)


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
@pytest.mark.parametrize(
    ("vstream_path", "hub_token"),
    [
        ("gcs_vstream_path", "hub_cloud_dev_token"),
        ("s3_vstream_path", "hub_cloud_dev_token"),
        ("hub_cloud_vstream_path", "hub_cloud_dev_token"),
    ],
    indirect=True,
)
def test_video_timestamps(vstream_path, hub_token):
    ds = hub.load(vstream_path, read_only=True, token=hub_token)

    with pytest.raises(ValueError):
        stamps = ds.mp4_videos[:2].timestamps

    stamps = ds.large_video[0, 12000:1199:-100].timestamps

    assert len(stamps) == 109

    # timestamp is 50, 24 fps video, 50 * 24 = 1200th frame
    assert stamps[-1] == 50

    # cover stepping without seeking
    stamps = ds.large_video[0, 1200:1300:2].timestamps

    assert len(stamps) == 50
    assert stamps[0] == 50


def test_video_exception(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc")
        with pytest.raises(Exception):
            stamps = ds.abc.timestamps


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_video_sequence(local_ds, video_paths):
    with local_ds as ds:
        ds.create_tensor("video_seq", htype="sequence[video]", sample_compression="mp4")
        ds.video_seq.append([hub.read(video_paths["mp4"][0]) for _ in range(3)])
        ds.video_seq.append([hub.read(video_paths["mp4"][1]) for _ in range(3)])

        with pytest.raises(ValueError):
            ds.video_seq[:2].timestamps

        with pytest.raises(ValueError):
            ds.video_seq[0].timestamps

        with pytest.raises(ValueError):
            ds.video_seq[0, :2].timestamps

        assert ds.video_seq[0][1, 5:10].timestamps.shape == (5,)


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_video_data(local_ds, video_paths):
    with local_ds as ds:
        ds.create_tensor("video", htype="video", sample_compression="mp4")
        for _ in range(3):
            ds.video.append(hub.read(video_paths["mp4"][0]))
        ds.video.append(hub.read(video_paths["mp4"][1]))

        data = ds.video[2].data()
        assert data["frames"].shape == (400, 360, 640, 3)
        assert data["timestamps"].shape == (400,)

        data = ds.video[:2, 4, :5, :5].data()
        assert data["frames"].shape == (2, 5, 5, 3)
        assert data["timestamps"].shape == (2, 1)

        data = ds.video[:2, 10:20].data()
        assert data["frames"].shape == (2, 10, 360, 640, 3)
        assert data["timestamps"].shape == (2, 10)

        with pytest.raises(DynamicTensorNumpyError):
            ds.video[2:].data()

        data = ds.video[2:].data(aslist=True)
        assert len(data["frames"]) == 2
        assert data["frames"][0].shape == ds.video[2].shape
        assert data["frames"][1].shape == ds.video[3].shape
        assert len(data["timestamps"]) == 2
        assert data["timestamps"][0].shape == (ds.video[2].shape[0],)
        assert data["timestamps"][1].shape == (ds.video[3].shape[0],)


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_linked_video_timestamps(local_ds):
    with local_ds as ds:
        ds.create_tensor("videos", htype="link[video]")
        ds.videos.append(
            hub.link(
                "http://commondatastorage.googleapis.com/gtv-videos-bucket/sample/ForBiggerBlazes.mp4",
                creds_key="ENV",
            )
        )
        ds.videos[0, 5:10].timestamps == np.array(
            [0.04170833, 0.08341666, 0.125125, 0.16683333, 0.20854166]
        )



from typing import Optional, Tuple, Union, List, Any
from hub.core.partial_sample import PartialSample
import numpy as np


def tiled(
    sample_shape: Tuple[int, ...],
    tile_shape: Optional[Tuple[int, ...]] = None,
    dtype: Union[str, np.dtype] = np.dtype("uint8"),
):
    """Allocates an empty sample of shape ``sample_shape``, broken into tiles of shape ``tile_shape`` (except for edge tiles).

    Example:

        >>> with ds:
        ...    ds.create_tensor("image", htype="image", sample_compression="png")
        ...    ds.image.append(hub.tiled(sample_shape=(1003, 1103, 3), tile_shape=(10, 10, 3)))
        ...    ds.image[0][-217:, :212, 1:] = np.random.randint(0, 256, (217, 212, 2), dtype=np.uint8)

    Args:
        sample_shape (Tuple[int, ...]): Full shape of the sample.
        tile_shape (Optional, Tuple[int, ...]): The sample will be will stored as tiles where each tile will have this shape (except edge tiles).
            If not specified, it will be computed such that each tile is close to half of the tensor's `max_chunk_size` (after compression).
        dtype (Union[str, np.dtype]): Dtype for the sample array. Default uint8.

    Returns:
        PartialSample: A PartialSample instance which can be appended to a Tensor.
    """
    return PartialSample(sample_shape=sample_shape, tile_shape=tile_shape, dtype=dtype)



from hub.experimental.convert_to_hub3 import dataset_to_hub3
from hub.util.bugout_reporter import hub_reporter


@hub_reporter.record_call
def query(dataset, query_string: str):
    """Returns a sliced hub.Dataset with given query results.

    It allows to run SQL like queries on dataset and extract results. Currently supported keywords are the following:

    +-------------------------------------------+
    | SELECT                                    |
    +-------------------------------------------+
    | FROM                                      |
    +-------------------------------------------+
    | CONTAINS                                  |
    +-------------------------------------------+
    | ORDER BY                                  |
    +-------------------------------------------+
    | GROUP BY                                  |
    +-------------------------------------------+
    | LIMIT                                     |
    +-------------------------------------------+
    | OFFSET                                    |
    +-------------------------------------------+
    | RANDOM() -> for shuffling query results   |
    +-------------------------------------------+


    Args:
        dataset: hub.Dataset object on which the query needs to be run
        query_string (str): An SQL string adjusted with new functionalities to run on the given hub.Dataset object


    Returns:
        Dataset: A hub.Dataset object.

    Examples:
        >>> import hub
        >>> from hub.experimental import query
        >>> ds = hub.load('hub://activeloop/fashion-mnist-train')
        >>> query_ds_train = query(ds_train, "select * shere labels != 5")

        >>> ds_train = hub.load('hub://activeloop/coco-train')
        >>> query_ds_train = query(ds_train, "(select * where contains(categories, 'car') limit 1000) union (select * where contains(categories, 'motorcycle') limit 1000)")
    """
    ds = dataset_to_hub3(dataset)
    dsv = ds.query(query_string)
    indexes = dsv.indexes
    return dataset[indexes]



import hub
import numpy as np
from typing import Dict, List
from hub.core.meta.encode.creds import CredsEncoder
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.meta.encode.chunk_id import ChunkIdEncoder
from hub.core.meta.encode.tile import TileEncoder
from hub.core.meta.encode.sequence import SequenceEncoder
from hub.core.storage.provider import StorageProvider
from hub.core.version_control.commit_chunk_set import CommitChunkSet
from hub.core.version_control.commit_diff import CommitDiff
from hub.util.keys import (
    get_creds_encoder_key,
    get_sequence_encoder_key,
    get_tensor_commit_chunk_set_key,
    get_tensor_commit_diff_key,
    get_tensor_meta_key,
    get_chunk_id_encoder_key,
    get_chunk_id_encoder_key,
    get_tensor_tile_encoder_key,
)
import posixpath


def merge_all_meta_info(
    target_ds, storage, generated_tensors, overwrite, all_num_samples, result
):
    merge_all_commit_diffs(
        result["commit_diffs"], target_ds, storage, overwrite, generated_tensors
    )
    merge_all_tile_encoders(
        result["tile_encoders"],
        all_num_samples,
        target_ds,
        storage,
        overwrite,
        generated_tensors,
    )
    merge_all_tensor_metas(
        result["tensor_metas"], target_ds, storage, overwrite, generated_tensors
    )
    merge_all_chunk_id_encoders(
        result["chunk_id_encoders"], target_ds, storage, overwrite, generated_tensors
    )
    merge_all_creds_encoders(
        result["creds_encoders"], target_ds, storage, overwrite, generated_tensors
    )
    merge_all_sequence_encoders(
        result["sequence_encoders"], target_ds, storage, overwrite, generated_tensors
    )
    if target_ds.commit_id is not None:
        merge_all_commit_chunk_sets(
            result["commit_chunk_sets"],
            target_ds,
            storage,
            overwrite,
            generated_tensors,
        )


def merge_all_tensor_metas(
    all_workers_tensor_metas: List[Dict[str, TensorMeta]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    """Merges tensor metas from all workers into a single one and stores it in target_ds."""
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)
        tensor_meta = None if overwrite else target_ds[rel_path].meta
        for current_worker_metas in all_workers_tensor_metas:
            current_meta = current_worker_metas[tensor]
            if tensor_meta is None:
                tensor_meta = current_meta
            else:
                combine_metas(tensor_meta, current_meta)
        meta_key = get_tensor_meta_key(tensor, commit_id)
        storage[meta_key] = tensor_meta.tobytes()  # type: ignore


def combine_metas(ds_tensor_meta: TensorMeta, worker_tensor_meta: TensorMeta) -> None:
    """Combines the dataset's tensor meta with a single worker's tensor meta."""
    # if tensor meta is empty, copy attributes from current_meta
    if len(ds_tensor_meta.max_shape) == 0 or ds_tensor_meta.dtype is None:
        ds_tensor_meta.set_dtype_str(worker_tensor_meta.dtype)
        if not ds_tensor_meta.htype and worker_tensor_meta.htype:
            ds_tensor_meta.set_htype(worker_tensor_meta.htype)
        ds_tensor_meta.update_length(worker_tensor_meta.length)
        ds_tensor_meta.update_shape_interval(worker_tensor_meta.max_shape)
        ds_tensor_meta.update_shape_interval(worker_tensor_meta.min_shape)
    # len of min_shape will be 0 if 0 outputs from worker
    elif len(worker_tensor_meta.min_shape) != 0:
        assert ds_tensor_meta.dtype == worker_tensor_meta.dtype
        assert ds_tensor_meta.htype == worker_tensor_meta.htype
        # TODO we can support this once we have ragged tensor support
        assert len(ds_tensor_meta.max_shape) == len(worker_tensor_meta.max_shape)
        assert len(ds_tensor_meta.min_shape) == len(worker_tensor_meta.min_shape)
        ds_tensor_meta.update_length(worker_tensor_meta.length)
        ds_tensor_meta.update_shape_interval(worker_tensor_meta.max_shape)
        ds_tensor_meta.update_shape_interval(worker_tensor_meta.min_shape)


def merge_all_chunk_id_encoders(
    all_workers_chunk_id_encoders: List[Dict[str, ChunkIdEncoder]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    """Merges chunk_id_encoders from all workers into a single one and stores it in target_ds."""
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)
        chunk_id_encoder = (
            None if overwrite else target_ds[rel_path].chunk_engine.chunk_id_encoder
        )
        for current_worker_chunk_id_encoders in all_workers_chunk_id_encoders:
            current_chunk_id_encoder = current_worker_chunk_id_encoders[tensor]
            if chunk_id_encoder is None:
                chunk_id_encoder = current_worker_chunk_id_encoders[tensor]
            else:
                combine_chunk_id_encoders(chunk_id_encoder, current_chunk_id_encoder)

        chunk_id_key = get_chunk_id_encoder_key(tensor, commit_id)
        storage[chunk_id_key] = chunk_id_encoder.tobytes()  # type: ignore


def combine_chunk_id_encoders(
    ds_chunk_id_encoder: ChunkIdEncoder,
    worker_chunk_id_encoder: ChunkIdEncoder,
) -> None:
    """Combines the dataset's chunk_id_encoder with a single worker's chunk_id_encoder."""
    encoded_ids = worker_chunk_id_encoder._encoded
    if not encoded_ids.flags.writeable:
        encoded_ids = encoded_ids.copy()
    if encoded_ids.size != 0:
        offset = ds_chunk_id_encoder.num_samples
        for encoded_id in encoded_ids:
            encoded_id[1] += offset
            if ds_chunk_id_encoder._encoded.size == 0:
                ds_chunk_id_encoder._encoded = np.reshape(encoded_id, (-1, 2))
            else:
                ds_chunk_id_encoder._encoded = np.vstack(
                    [ds_chunk_id_encoder._encoded, encoded_id]
                )


def merge_all_tile_encoders(
    all_workers_tile_encoders: List[Dict[str, TileEncoder]],
    all_num_samples: List[Dict[str, int]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)
        chunk_engine = target_ds[rel_path].chunk_engine
        offset = 0 if overwrite else chunk_engine.num_samples
        tile_encoder = None if overwrite else chunk_engine.tile_encoder
        for i, current_worker_tile_encoder in enumerate(all_workers_tile_encoders):
            current_tile_encoder = current_worker_tile_encoder[tensor]
            if tile_encoder is None:
                tile_encoder = current_tile_encoder
            else:
                combine_tile_encoders(tile_encoder, current_tile_encoder, offset)
            offset += all_num_samples[i][tensor]
        tile_key = get_tensor_tile_encoder_key(tensor, commit_id)
        storage[tile_key] = tile_encoder.tobytes()  # type: ignore
    target_ds.flush()


def combine_tile_encoders(
    ds_tile_encoder: TileEncoder, worker_tile_encoder: TileEncoder, offset: int
) -> None:
    """Combines the dataset's tile_encoder with a single worker's tile_encoder."""

    if len(worker_tile_encoder.entries) != 0:
        for sample_index in worker_tile_encoder.entries.keys():
            new_sample_index = int(sample_index) + offset

            if new_sample_index in ds_tile_encoder.entries:
                raise ValueError(
                    f"Sample index {new_sample_index} already exists inside `ds_tile_encoder`. Keys={ds_tile_encoder.entries}"
                )

            ds_tile_encoder.entries[new_sample_index] = worker_tile_encoder.entries[
                sample_index
            ]


def merge_all_commit_chunk_sets(
    all_workers_commit_chunk_sets: List[Dict[str, CommitChunkSet]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    """Merges commit_chunk_sets from all workers into a single one and stores it in target_ds."""
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)
        commit_chunk_set = (
            None if overwrite else target_ds[rel_path].chunk_engine.commit_chunk_set
        )
        for current_worker_commit_chunk_set in all_workers_commit_chunk_sets:
            current_commit_chunk_set = current_worker_commit_chunk_set[tensor]
            if commit_chunk_set is None:
                commit_chunk_set = current_commit_chunk_set
            else:
                combine_commit_chunk_sets(commit_chunk_set, current_commit_chunk_set)

        commit_chunk_key = get_tensor_commit_chunk_set_key(tensor, commit_id)
        storage[commit_chunk_key] = commit_chunk_set.tobytes()  # type: ignore


def combine_commit_chunk_sets(
    ds_commit_chunk_set: CommitChunkSet,
    worker_commit_chunk_set: CommitChunkSet,
) -> None:
    """Combines the dataset's commit_chunk_set with a single worker's commit_chunk_set."""
    ds_commit_chunk_set.chunks.update(worker_commit_chunk_set.chunks)


def merge_all_commit_diffs(
    all_workers_commit_diffs: List[Dict[str, CommitDiff]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    """Merges commit_diffs from all workers into a single one and stores it in target_ds."""
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)  # type: ignore
        commit_diff = None if overwrite else target_ds[rel_path].chunk_engine.commit_diff  # type: ignore
        for current_worker_commit_diffs in all_workers_commit_diffs:
            current_commit_diff = current_worker_commit_diffs[tensor]
            if commit_diff is None:
                commit_diff = current_commit_diff
                commit_diff.transform_data()
            else:
                combine_commit_diffs(commit_diff, current_commit_diff)

        commit_chunk_key = get_tensor_commit_diff_key(tensor, commit_id)
        storage[commit_chunk_key] = commit_diff.tobytes()  # type: ignore


def combine_commit_diffs(
    ds_commit_diff: CommitDiff, worker_commit_diff: CommitDiff
) -> None:
    """Combines the dataset's commit_diff with a single worker's commit_diff."""
    ds_commit_diff.add_data(worker_commit_diff.num_samples_added)


def merge_all_creds_encoders(
    all_workers_creds_encoders: List[Dict[str, CredsEncoder]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)
        actual_tensor = target_ds[rel_path]
        if not actual_tensor.is_link:
            continue

        creds_encoder = None if overwrite else actual_tensor.chunk_engine.creds_encoder
        for current_worker_creds_encoder in all_workers_creds_encoders:
            current_creds_encoder = current_worker_creds_encoder[tensor]
            if creds_encoder is None:
                creds_encoder = current_creds_encoder
            else:
                combine_creds_encoders(creds_encoder, current_creds_encoder)

        creds_key = get_creds_encoder_key(tensor, commit_id)
        storage[creds_key] = creds_encoder.tobytes()  # type: ignore


def combine_creds_encoders(
    ds_creds_encoder: CredsEncoder, worker_creds_encoder: CredsEncoder
) -> None:
    """Combines the dataset's creds_encoder with a single worker's creds_encoder."""
    arr = worker_creds_encoder.array
    num_entries = len(arr)
    last_index = -1
    for i in range(num_entries):
        next_last_index = arr[i][1]
        num_samples = next_last_index - last_index
        ds_creds_encoder.register_samples((arr[i][0],), num_samples)
        last_index = next_last_index


def merge_all_sequence_encoders(
    all_workers_sequence_encoders: List[Dict[str, SequenceEncoder]],
    target_ds: hub.Dataset,
    storage: StorageProvider,
    overwrite: bool,
    tensors: List[str],
) -> None:
    commit_id = target_ds.version_state["commit_id"]
    for tensor in tensors:
        rel_path = posixpath.relpath(tensor, target_ds.group_index)
        actual_tensor = target_ds[rel_path]
        if not actual_tensor.is_sequence:
            continue
        sequence_encoder = (
            None if overwrite else actual_tensor.chunk_engine.sequence_encoder
        )
        for current_worker_sequence_encoder in all_workers_sequence_encoders:
            current_sequence_encoder = current_worker_sequence_encoder[tensor]
            if sequence_encoder is None:
                sequence_encoder = current_sequence_encoder
            else:
                combine_sequence_encoders(sequence_encoder, current_sequence_encoder)

        sequence_key = get_sequence_encoder_key(tensor, commit_id)
        storage[sequence_key] = sequence_encoder.tobytes()  # type: ignore


def combine_sequence_encoders(
    ds_sequence_encoder: SequenceEncoder, worker_sequence_encoder: SequenceEncoder
) -> None:
    """Combines the dataset's sequence_encoder with a single worker's sequence_encoder."""
    arr = worker_sequence_encoder.array
    last_index = -1
    for i in range(len(arr)):
        next_last_index = arr[i][2]
        ds_sequence_encoder.register_samples(arr[i][0], next_last_index - last_index)
        last_index = next_last_index



from hub.constants import MB, PARTIAL_NUM_SAMPLES
from hub.core.chunk.uncompressed_chunk import UncompressedChunk
import numpy as np
import pytest

import hub
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.sample import Sample  # type: ignore
from hub.core.tiling.deserialize import np_list_to_sample
from hub.core.tiling.sample_tiles import SampleTiles


common_args = {
    "min_chunk_size": 1 * MB,
    "max_chunk_size": 2 * MB,
    "tiling_threshold": 1 * MB,
    "compression": None,
}


def create_tensor_meta():
    tensor_meta = TensorMeta()
    tensor_meta.dtype = "float64"
    tensor_meta.max_shape = None
    tensor_meta.min_shape = None
    tensor_meta.htype = None
    tensor_meta.length = 0
    return tensor_meta


def test_read_write_sequence():
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    dtype = tensor_meta.dtype
    data_in = [np.random.rand(125, 125).astype(dtype) for _ in range(10)]
    while data_in:
        chunk = UncompressedChunk(**common_args)
        num_samples = int(chunk.extend_if_has_space(data_in))
        data_out = [chunk.read_sample(i) for i in range(num_samples)]
        np.testing.assert_array_equal(data_out, data_in[:num_samples])
        data_in = data_in[num_samples:]


def test_read_write_sequence_big(cat_path):
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    dtype = tensor_meta.dtype
    data_in = []
    for i in range(50):
        if i % 10 == 0:
            data_in.append(np.random.rand(751, 750, 3).astype(dtype))
        elif i % 3 == 0:
            data_in.append(hub.read(cat_path))
        else:
            data_in.append(np.random.rand(125, 125, 3).astype(dtype))
    data_in2 = data_in.copy()
    tiles = []
    original_length = len(data_in)

    while data_in:
        chunk = UncompressedChunk(**common_args)
        num_samples = chunk.extend_if_has_space(data_in)
        if num_samples == PARTIAL_NUM_SAMPLES:
            tiles.append(chunk.read_sample(0, is_tile=True))
            sample = data_in[0]
            assert isinstance(sample, SampleTiles)
            if sample.is_last_write:
                current_length = len(data_in)
                index = original_length - current_length
                full_data_out = np_list_to_sample(
                    tiles,
                    sample.sample_shape,
                    sample.tile_shape,
                    sample.layout_shape,
                    dtype,
                )
                np.testing.assert_array_equal(full_data_out, data_in2[index])
                data_in = data_in[1:]
                tiles = []
        elif num_samples > 0:
            data_out = [chunk.read_sample(i) for i in range(num_samples)]
            for i, item in enumerate(data_out):
                if isinstance(item, Sample):
                    item = item.array
                np.testing.assert_array_equal(item, data_in[i])
            data_in = data_in[num_samples:]


def test_read_write_numpy():
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    dtype = tensor_meta.dtype
    data_in = np.random.rand(10, 125, 125).astype(dtype)
    while len(data_in) > 0:
        chunk = UncompressedChunk(**common_args)
        num_samples = int(chunk.extend_if_has_space(data_in))
        data_out = np.array([chunk.read_sample(i) for i in range(num_samples)])
        if num_samples > 0:
            np.testing.assert_array_equal(data_out, data_in[:num_samples])
        data_in = data_in[num_samples:]


def test_read_write_numpy_big():
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    dtype = tensor_meta.dtype
    data_in = np.random.rand(2, 750, 750, 3).astype(dtype)
    prev_num_samples = None
    with pytest.raises(ValueError):
        while len(data_in) > 0:
            chunk = UncompressedChunk(**common_args)
            num_samples = int(chunk.extend_if_has_space(data_in))
            if num_samples == 0 and prev_num_samples == 0:
                raise ValueError(
                    "Unexpected, bigger numpy arrays should be sent as sequence to chunk"
                )
            data_out = np.array([chunk.read_sample(i) for i in range(num_samples)])
            if num_samples > 0:
                np.testing.assert_array_equal(data_out, data_in[:num_samples])
            data_in = data_in[num_samples:]
            prev_num_samples = num_samples


def test_update():
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    dtype = tensor_meta.dtype
    data_in = np.random.rand(7, 125, 125).astype(dtype)
    chunk = UncompressedChunk(**common_args)
    chunk.extend_if_has_space(data_in)

    data_out = np.array([chunk.read_sample(i) for i in range(7)])
    np.testing.assert_array_equal(data_out, data_in)

    data_3 = np.random.rand(175, 175).astype(dtype)
    data_5 = np.random.rand(375, 375).astype(dtype)

    chunk.update_sample(3, data_3)
    chunk.update_sample(5, data_5)
    for i in range(7):
        if i == 3:
            np.testing.assert_array_equal(chunk.read_sample(i), data_3)
        elif i == 5:
            np.testing.assert_array_equal(chunk.read_sample(i), data_5)
        else:
            np.testing.assert_array_equal(chunk.read_sample(i), data_in[i])



from typing import Dict, Optional, Any
from hub.util.logging import log_visualizer_link


class ViewEntry:
    """Represents a view saved inside a dataset."""

    def __init__(
        self, info: Dict, dataset, source_dataset=None, external: bool = False
    ):
        self.info = info
        self._ds = dataset
        self._src_ds = source_dataset if external else dataset
        self._external = external

    def __getitem__(self, key: str):
        return self.info[key]

    def get(self, key: str, default: Optional[Any] = None):
        return self.info.get(key, default)

    @property
    def id(self) -> str:
        """Returns id of the view."""
        return self.info["id"].split("]")[-1]

    @property
    def query(self) -> Optional[str]:
        return self.info.get("query")

    @property
    def message(self) -> str:
        """Returns the message with which the view was saved."""
        return self.info.get("message", "")

    def __str__(self):
        return f"View(id='{self.id}', message='{self.message}', virtual={self.virtual})"

    __repr__ = __str__

    @property
    def virtual(self) -> bool:
        return self.info["virtual-datasource"]

    def load(self, verbose=True):
        """Loads the view and returns the :class:`~hub.core.dataset.Dataset`.

        Args:
            verbose (bool): If ``True``, logs will be printed. Defaults to ``True``.

        Returns:
            Dataset: Loaded dataset view.
        """
        ds = self._ds._sub_ds(
            ".queries/" + (self.info.get("path") or self.info["id"]),
            lock=False,
            verbose=False,
        )
        sub_ds_path = ds.path
        if self.virtual:
            ds = ds._get_view(inherit_creds=not self._external)
        ds._view_entry = self
        if verbose:
            log_visualizer_link(sub_ds_path, source_ds_url=self.info["source-dataset"])
        return ds

    def optimize(
        self, unlink=True, num_workers=0, scheduler="threaded", progressbar=True
    ):
        """Optimizes the dataset view by copying and rechunking the required data. This is necessary to achieve fast streaming
        speeds when training models using the dataset view. The optimization process will take some time, depending on
        the size of the data.

        Example:

            >>> # save view
            >>> ds[:10].save_view(view_id="first_10")
            >>> # optimize view
            >>> ds.get_view("first_10").optimize()
            >>> # load optimized view
            >>> ds.load_view("first_10")

        Args:
            unlink (bool): - If ``True``, this unlinks linked tensors (if any) by copying data from the links to the view.
                    - This does not apply to linked videos. Set ``hub.constants._UNLINK_VIDEOS`` to ``True`` to change this behavior.
            num_workers (int): Number of workers to be used for the optimization process. Defaults to 0.
            scheduler (str): The scheduler to be used for optimization. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
                Only applicable if ``optimize=True``. Defaults to 'threaded'.
            progressbar (bool): Whether to display a progressbar.

        Returns:
            :class:`ViewEntry`
        """
        self.info = self._ds._optimize_saved_view(
            self.info["id"],
            external=self._external,
            unlink=unlink,
            num_workers=num_workers,
            scheduler=scheduler,
            progressbar=progressbar,
        )
        return self

    @property
    def source_dataset_path(self) -> str:
        return self.info["source-dataset"]

    def delete(self):
        """Deletes the view."""
        self._ds.delete_view(id=self.info["id"])



import numpy as np
from typing import Tuple, Union, List, Optional


def get_tile_shape(
    sample_shape: Tuple[int, ...],
    sample_size: Optional[float] = None,
    chunk_size: int = 16 * 2**20,
    exclude_axes: Optional[Union[int, List[int]]] = None,
) -> Tuple[int, ...]:
    """
    Get tile shape for a given sample shape that will fit in chunk_size

    Args:
        sample_shape: Shape of the sample
        sample_size: Size of the compressed sample in bytes
        chunk_size: Expected size of a compressed tile in bytes
        exclude_axes: Dimensions to be excluded from tiling. (2 for RGB images)

    Returns:
        Tile shape

    Raises:
        ValueError: If the chunk_size is too small
    """
    ratio = sample_size / chunk_size  # type: ignore
    sample_shape = np.array(sample_shape, dtype=np.float32)  # type: ignore
    if isinstance(exclude_axes, int):
        exclude_axes = [exclude_axes]
    elif exclude_axes is None:
        exclude_axes = []
    elif not isinstance(exclude_axes, list):
        # should be a list for numpy advanced indexing
        exclude_axes = list(exclude_axes)
    sample_shape_masked = sample_shape.copy()  # type: ignore
    sample_shape_masked[exclude_axes] = 0
    while ratio > 1:
        idx = np.argmax(sample_shape_masked)
        val = sample_shape_masked[idx : idx + 1]  # type: ignore
        if val < 2:
            raise ValueError(f"Chunk size is too small: {chunk_size} bytes")
        val /= 2
        ratio /= 2
    sample_shape_masked[exclude_axes] = sample_shape[exclude_axes]  # type: ignore
    arr = np.ceil(sample_shape_masked)
    # convert arr to a tuple of python integers
    return tuple(int(x) for x in arr)



import numpy as np
from typing import List, Tuple, Union, Optional
from hub.core.chunk.base_chunk import BaseChunk
from hub.core.meta.encode.tile import TileEncoder


def coalesce_tiles(
    tiles: np.ndarray,
    tile_shape: Tuple[int, ...],
    sample_shape: Optional[Tuple[int, ...]],
    dtype: Union[str, np.dtype],
) -> np.ndarray:
    """Coalesce tiles into a single array of shape `sample_shape`.
    Args:
        tiles (np.ndarray): numpy object array of tiles.
        tile_shape (Tuple[int, ...]): Tile shape. Corner tiles may be smaller than this.
        sample_shape (Optional, Tuple[int, ...]): Shape of the output array. The sum of all actual tile shapes are expected to be equal to this.
        dtype (Union[str, np.dtype]): Dtype of the output array. Should match dtype of tiles.
    Raises:
        TypeError: If `tiles` is not deserialized.
    Returns:
        np.ndarray: Sample array from tiles.
    """
    ndim = tiles.ndim
    sample_shape = sample_shape or tuple(  # Infer sample shape from tile shapes
        sum(
            tile.shape[i]
            for tile in tiles[tuple(slice(None) if j == i else 0 for j in range(ndim))]  # type: ignore
        )
        for i in range(ndim)
    )
    sample = np.empty(sample_shape, dtype=dtype)
    if tiles.size <= 0:
        return sample

    for tile_coords, tile in np.ndenumerate(tiles):
        low = np.multiply(tile_coords, tile_shape)
        high = low + tile.shape
        idx = tuple(slice(l, h) for l, h in zip(low, high))
        view = sample[idx]
        view[:] = tile
    return sample


def combine_chunks(
    chunks: List[BaseChunk],
    sample_index: int,
    tile_encoder: TileEncoder,
) -> np.ndarray:
    dtype = chunks[0].dtype
    shape = tile_encoder.get_sample_shape(sample_index)
    tile_shape = tile_encoder.get_tile_shape(sample_index)
    layout_shape = tile_encoder.get_tile_layout_shape(sample_index)

    # index is always 0 within a chunk for tiled samples
    tiled_arrays = [chunk.read_sample(0, is_tile=True) for chunk in chunks]
    return np_list_to_sample(tiled_arrays, shape, tile_shape, layout_shape, dtype)


def np_list_to_sample(
    tiled_arrays: List[np.ndarray], shape, tile_shape, layout_shape, dtype
) -> np.ndarray:
    num_tiles = len(tiled_arrays)
    tiles = np.empty((num_tiles,), dtype=object)
    tiles[:] = tiled_arrays[:]
    tiles = np.reshape(tiles, layout_shape)
    return coalesce_tiles(tiles, tile_shape, shape, dtype)


def translate_slices(
    slices: List[Union[slice, int, List[int]]],
    sample_shape: Tuple[int, ...],
    tile_shape: Tuple[int, ...],
) -> Tuple[Tuple, Tuple]:
    """Translates slices from sample space to tile space
    Args:
        sample_shape (Tuple[int, ...]): Sample shape.
        tile_shape (Tuple[int, ...]): Tile shape.
    Raises:
        NotImplementedError: For stepping slices
    """
    tiles_index: List[slice] = []
    sample_index: List[Union[int, slice, List[int]]] = []
    for i, s in enumerate(slices):
        if isinstance(s, int):
            if s < 0:
                s += sample_shape[i]
            ts = s // tile_shape[i]
            tiles_index.append(slice(ts, ts + 1))
            sample_index.append(s % tile_shape[i])
        elif isinstance(s, list):
            s = [x + sample_shape[i] if x < 0 else x for x in s]
            mn, mx = min(s), max(s)
            tiles_index.append(slice(mn // tile_shape[i], mx // tile_shape[i] + 1))
            sample_index.append([x - mn for x in s])
        elif isinstance(s, slice):
            start, stop, step = s.start, s.stop, s.step
            if start is None:
                start = 0
            elif start < 0:
                start += sample_shape[i]
            if stop is None:
                stop = sample_shape[i]
            elif stop < 0:
                stop += sample_shape[i]
            else:
                stop = min(stop, sample_shape[i])
            if step not in (1, None):
                raise NotImplementedError(
                    "Stepped indexing for tiled samples is not supported yet."
                )
            ts = slice(start // tile_shape[i], (stop - 1) // tile_shape[i] + 1)  # type: ignore
            tiles_index.append(ts)  # type: ignore
            offset = ts.start * tile_shape[i]  # type: ignore
            sample_index.append(slice(start - offset, stop - offset))  # type: ignore
    return tuple(tiles_index), tuple(sample_index)



import hub
import numpy as np
import pytest
from unittest.mock import patch


all_non_image_compressions = pytest.mark.parametrize(
    "compression",
    [
        {"sample_compression": None},
        {"sample_compression": "lz4"},
        {"chunk_compression": "lz4"},
    ],
)


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
@all_non_image_compressions
@pytest.mark.parametrize("insert_first", [True, False])
def test_insertion_array(memory_ds, compression, insert_first):
    with memory_ds as ds:
        ds.create_tensor("abc", **compression)
        first = np.random.rand(100, 100, 3)
        tenth = np.random.rand(100, 100, 3)
        empty_sample = np.random.rand(0, 0, 0)
        if insert_first:
            ds.abc.append(first)
        ds.abc[10] = tenth

        if insert_first:
            np.testing.assert_array_equal(ds.abc[0].numpy(), first)
        else:
            np.testing.assert_array_equal(ds.abc[0].numpy(), empty_sample)

        for i in range(1, 10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), empty_sample)
        np.testing.assert_array_equal(ds.abc[10].numpy(), tenth)


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
@pytest.mark.parametrize("sample_compression", ["png", "jpeg"])
@pytest.mark.parametrize("insert_first", [True, False])
def test_insertion_array_img_compressed(memory_ds, sample_compression, insert_first):
    with memory_ds as ds:
        ds.create_tensor("abc", sample_compression=sample_compression)
        first = np.random.randint(0, 256, (200, 300, 3), dtype=np.uint8)
        tenth = np.random.randint(0, 256, (100, 100, 3), dtype=np.uint8)
        empty_sample = np.random.randint(0, 256, (0, 0, 0), dtype=np.uint8)
        if insert_first:
            ds.abc.append(first)
        ds.abc[10] = tenth

        if insert_first:
            if sample_compression == "png":
                np.testing.assert_array_equal(ds.abc[0].numpy(), first)
            else:
                assert ds.abc[0].numpy().shape == first.shape
        else:
            np.testing.assert_array_equal(ds.abc[0].numpy(), empty_sample)

        for i in range(1, 10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), empty_sample)

        if sample_compression == "png":
            np.testing.assert_array_equal(ds.abc[10].numpy(), tenth)
        else:
            assert ds.abc[10].numpy().shape == tenth.shape


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
@all_non_image_compressions
@pytest.mark.parametrize("insert_first", [True, False])
def test_insertion_json(memory_ds, compression, insert_first):
    with memory_ds as ds:
        ds.create_tensor("abc", **compression)
        first = {"a": 1, "b": 2}
        tenth = {"a": 3, "b": 4}
        empty_sample = {}
        if insert_first:
            ds.abc.append(first)
        ds.abc[10] = tenth

        if insert_first:
            assert ds.abc[0].numpy()[0] == first
        else:
            assert ds.abc[0].numpy()[0] == empty_sample

        for i in range(1, 10):
            assert ds.abc[i].numpy() == empty_sample

        assert ds.abc[10].numpy() == tenth


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
@all_non_image_compressions
@pytest.mark.parametrize("insert_first", [True, False])
def test_insertion_text(memory_ds, compression, insert_first):
    with memory_ds as ds:
        ds.create_tensor("abc", **compression)
        first = "hi"
        tenth = "if ur reading this ur a nerd"
        empty_sample = ""
        if insert_first:
            ds.abc.append(first)
        ds.abc[10] = tenth

        if insert_first:
            assert ds.abc[0].numpy()[0] == first
        else:
            assert ds.abc[0].numpy()[0] == empty_sample

        for i in range(1, 10):
            assert ds.abc[i].numpy() == empty_sample

        assert ds.abc[10].numpy() == tenth


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
@all_non_image_compressions
@pytest.mark.parametrize("insert_first", [True, False])
def test_insertion_list(memory_ds, compression, insert_first):
    with memory_ds as ds:
        ds.create_tensor("abc", **compression, htype="list")
        first = [1, 2, 3]
        tenth = [4, 5, 6]
        empty_sample = np.array([], dtype="object")
        if insert_first:
            ds.abc.append(first)
        ds.abc[10] = tenth

        if insert_first:
            np.testing.assert_array_equal(ds.abc[0].numpy(), np.array(first))
        else:
            np.testing.assert_array_equal(ds.abc[0].numpy(), empty_sample)

        for i in range(1, 10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), empty_sample)

        np.testing.assert_array_equal(ds.abc[10].numpy(), np.array(tenth))


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
def test_updation_bug(memory_ds):
    with memory_ds as ds:
        labels = ds.create_tensor("labels", "class_label")
        labels[0] = [0, 1]
        np.testing.assert_array_equal(labels[0].numpy(), [0, 1])
        labels[0] = [1, 2]
        np.testing.assert_array_equal(labels[0].numpy(), [1, 2])


@patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
@pytest.mark.parametrize("insert_first", [False, True])
def test_insertion_link(memory_ds, insert_first, cat_path, flower_path):
    with memory_ds as ds:
        ds.create_tensor("abc", htype="link[image]")
        first = hub.link(cat_path)
        tenth = hub.link(flower_path)
        empty_sample = np.ones((0,))
        if insert_first:
            ds.abc.append(first)
        ds.abc[10] = tenth

        if insert_first:
            assert ds.abc[0].numpy().shape == (900, 900, 3)
        else:
            np.testing.assert_array_equal(ds.abc[0].numpy(), empty_sample)

        for i in range(1, 10):
            np.testing.assert_array_equal(ds.abc[i].numpy(), empty_sample)

        assert ds.abc[10].numpy().shape == (513, 464, 4)


# removing this test until we fix sequence updates
# @patch("hub.constants._ENABLE_RANDOM_ASSIGNMENT", True)
# @pytest.mark.parametrize("insert_first", [True])
# def test_insertion_sequence(memory_ds, insert_first):
#     with memory_ds as ds:
#         ds.create_tensor("abc", htype="sequence")
#         first = [[1, 2, 3, 4], [5, 6, 7, 8]]
#         tenth = [[9, 10, 11, 12], [13, 14, 15, 16], [17, 18, 19, 20]]
#         empty_sample = np.ones((0, 0))
#         if insert_first:
#             ds.abc.append(first)
#         ds.abc[10] = tenth
#         ds.abc[0].numpy()

#         if insert_first:
#             np.testing.assert_array_equal(ds.abc[0].numpy(), first)
#         else:
#             np.testing.assert_array_equal(ds.abc[0].numpy(), empty_sample)

#         for i in range(1, 10):
#             np.testing.assert_array_equal(ds.abc[i].numpy(), empty_sample)

#         np.testing.assert_array_equal(ds.abc[10].numpy(), tenth)



from typing import Callable, Dict, List, Optional
from hub.util.iterable_ordered_dict import IterableOrderedDict
import numpy as np


def collate_fn(batch):
    import torch

    elem = batch[0]

    if isinstance(elem, IterableOrderedDict):
        return IterableOrderedDict(
            (key, collate_fn([d[key] for d in batch])) for key in elem.keys()
        )

    if isinstance(elem, np.ndarray) and elem.size > 0 and isinstance(elem[0], str):
        batch = [it[0] for it in batch]
    return torch.utils.data._utils.collate.default_collate(batch)


def convert_fn(data):
    import torch

    if isinstance(data, IterableOrderedDict):
        return IterableOrderedDict((k, convert_fn(v)) for k, v in data.items())
    if isinstance(data, np.ndarray) and data.size > 0 and isinstance(data[0], str):
        data = data[0]

    return torch.utils.data._utils.collate.default_convert(data)


class PytorchTransformFunction:
    def __init__(
        self,
        transform_dict: Optional[Dict[str, Optional[Callable]]] = None,
        composite_transform: Optional[Callable] = None,
    ) -> None:
        self.composite_transform = composite_transform
        self.transform_dict = transform_dict

    def __call__(self, data_in: Dict) -> Dict:
        if self.composite_transform is not None:
            return self.composite_transform(data_in)
        elif self.transform_dict is not None:
            data_out = {}
            for tensor, fn in self.transform_dict.items():
                value = data_in[tensor]
                data_out[tensor] = value if fn is None else fn(value)
            data_out = IterableOrderedDict(data_out)
            return data_out
        return data_in



from abc import ABC, abstractmethod
import json
from typing import Any, Dict


class HubMemoryObject(ABC):
    def __init__(self):
        self.is_dirty = True

    @property
    @abstractmethod
    def nbytes(self):
        """Returns the number of bytes in the object."""

    def __getstate__(self) -> Dict[str, Any]:
        return self.__dict__

    def __setstate__(self, state: Dict[str, Any]):
        self.__dict__.update(state)

    def tobytes(self) -> bytes:
        d = {str(k): v for k, v in self.__getstate__().items()}
        return bytes(json.dumps(d, sort_keys=True, indent=4), "utf-8")

    @classmethod
    def frombuffer(cls, buffer: bytes):
        instance = cls()
        if len(buffer) > 0:
            instance.__setstate__(json.loads(buffer))
            instance.is_dirty = False
            return instance
        raise BufferError("Unable to instantiate the object as the buffer was empty.")



import numpy as np
import pytest
from hub.util.exceptions import MemoryDatasetCanNotBePickledError
import pickle


@pytest.mark.parametrize(
    "ds",
    ["memory_ds", "local_ds", "s3_ds", "gcs_ds", "hub_cloud_ds"],
    indirect=True,
)
def test_dataset(ds):
    if ds.path.startswith("mem://"):
        with pytest.raises(MemoryDatasetCanNotBePickledError):
            pickle.dumps(ds)
        return

    with ds:
        ds.create_tensor("image", htype="image", sample_compression="jpeg")
        ds.create_tensor("label")
        for i in range(10):
            ds.image.append(
                i * np.ones(((i + 1) * 20, (i + 1) * 20, 3), dtype=np.uint8)
            )

        for i in range(5):
            ds.label.append(i)

    pickled_ds = pickle.dumps(ds)
    unpickled_ds = pickle.loads(pickled_ds)
    assert len(unpickled_ds.image) == len(ds.image)
    assert len(unpickled_ds.label) == len(ds.label)
    assert unpickled_ds.tensors.keys() == ds.tensors.keys()
    assert unpickled_ds.index.values[0].value == ds.index.values[0].value
    assert unpickled_ds.meta.version == ds.meta.version

    for i in range(10):
        np.testing.assert_array_equal(
            ds.image[i].numpy(),
            (i * np.ones(((i + 1) * 20, (i + 1) * 20, 3), dtype=np.uint8)),
        )
        np.testing.assert_array_equal(
            ds.image[i].numpy(), unpickled_ds.image[i].numpy()
        )
    for i in range(5):
        np.testing.assert_array_equal(ds.label[i].numpy(), i)
        np.testing.assert_array_equal(
            ds.label[i].numpy(), unpickled_ds.label[i].numpy()
        )



import hub
from tokenize import tokenize, TokenError
from io import BytesIO
from typing import Any, Dict, List


def _token_obj_to_dict(token):
    if token.end[0] != 1:
        raise NotImplementedError("Only single line queries are supported yet.")
    return {
        "string": token.string,
        "start": token.start[1],
        "end": token.end[1],
        #  Note: type will be filled in by _tokenize()
    }


_TENSOR_PROPERTIES = set(
    (
        "min",
        "max",
        "mean",
        "shape",
        "size",
    )
)

_TENSOR_METHODS = set(("contains",))


_PYTHON_KEYWORDS = set(
    (
        "in",
        "is",
        "and",
        "or",
        "not",
        "if",
        "else",
        "for",
    )
)

_PYTHON_CONSTANTS = set(
    (
        "True",
        "False",
        "None",
    )
)


def _tokenize(s: str):
    return list(tokenize(BytesIO(s.encode("utf-8")).readline))[1:-2]


def _filter_hidden_tensors(tensors):
    tensor_keys = set(tensors.keys())
    for key in tensor_keys:
        tensor = tensors[key]
        meta = tensor.meta
        if hasattr(meta, "hidden") and tensor.meta.hidden:
            tensors.pop(key)
    return tensors


def _parse(s: str, ds: hub.Dataset) -> List[dict]:
    """Tokenizes a query string and assigns a token type to each token.

    Args:
        s (str): The query string
        ds (hub.Dataset): The dataset against which the query is run

    Returns:
        List of "tokens", each token is a `dict` with following fields:
            "string": the actual token string.
            "start" and "end": indices for the token in the query string.
            "type": Token type.

    """
    pytokens = _tokenize(s)
    tensors = _filter_hidden_tensors(ds._ungrouped_tensors)
    groups = set(ds._groups_filtered)
    hubtokens: List[Dict[str, Any]] = []
    group_in_progress = None
    for i, t in enumerate(pytokens):
        ht = _token_obj_to_dict(t)
        ht["type"] = "UNKNOWN"
        ts = t.string
        if t.type == 1:  # (NAME)
            if group_in_progress:
                if ts in _filter_hidden_tensors(group_in_progress._ungrouped_tensors):
                    ht["type"] = "TENSOR"
                    group_in_progress = None
                elif ts in group_in_progress._groups_filtered:
                    ht["type"] = "GROUP"
                    group_in_progress = group_in_progress[ts]
                else:
                    group_in_progress = None
            if ts in _PYTHON_KEYWORDS:
                ht["type"] = "KEYWORD"
            elif ts in _PYTHON_CONSTANTS:
                ht["type"] = "CONSTANT"
            elif ts in tensors:
                ht["type"] = "TENSOR"
            elif ts in groups:
                ht["type"] = "GROUP"
                group_in_progress = ds[ts]
            elif ts in _TENSOR_PROPERTIES:
                if (
                    i >= 2
                    and hubtokens[-1]["string"] == "."
                    and hubtokens[-2]["type"] == "TENSOR"
                ):
                    ht["type"] = "PROPERTY"
                else:
                    pass  # Syntax error
            elif ts in _TENSOR_METHODS:
                if (
                    i >= 2
                    and hubtokens[-1]["string"] == "."
                    and hubtokens[-2]["type"] == "TENSOR"
                ):
                    ht["type"] = "METHOD"
                else:
                    pass  # Syntax error
        elif t.type == 2:
            ht["type"] = "NUMBER"
            group_in_progress = None
        elif t.type == 3:
            ht["type"] = "STRING"
            group_in_progress = None
        elif t.type == 53:  # OP
            if ts != ".":
                group_in_progress = None
            elif i >= 2 and hubtokens[-1]["string"] == ".":
                group_in_progress = None  # Syntax error
            ht["type"] = "OP"
        hubtokens.append(ht)
    return hubtokens


def _parse_no_fail(s: str, ds: hub.Dataset):
    """Python tokenizer can fail for some partial queries such as those with trailing "(".
    This method supresses such errors.
    """
    try:
        return _parse(s, ds)
    except TokenError:
        return _parse(s[:-1], ds)


def _sort_suggestions(s: List[dict]) -> None:
    """Sorts a list of autocomplete suggestions alphabetically."""
    s.sort(key=lambda s: s["string"])


def _initial_suggestions(ds: hub.Dataset):
    """Suggestions for empty string query."""
    tensors = [
        {"string": k, "type": "TENSOR"}
        for k in _filter_hidden_tensors(ds._ungrouped_tensors)
    ]
    _sort_suggestions(tensors)
    groups = [{"string": k, "type": "GROUP"} for k in ds._groups_filtered]
    _sort_suggestions(groups)
    groups.sort(key=lambda s: s["string"])
    return tensors + groups


def _tensor_suggestions(ds, tensor):
    """Suggestions when last token of a query is a tensor name."""
    methods = [{"string": m, "type": "METHOD"} for m in _TENSOR_METHODS]
    _sort_suggestions(methods)
    properties = [{"string": p, "type": "PROPERTY"} for p in _TENSOR_PROPERTIES]
    _sort_suggestions(properties)
    return methods + properties


def _const_suggestions():
    """Suggest True, False and None."""
    return [{"string": str(v), "type": "CONSTANT"} for v in (True, False, None)]


def _group_suggestions(ds, group):
    """Suggestions when last token of a query is a group name."""
    return [
        {"string": k, "type": "TENSOR"}
        for k in _filter_hidden_tensors(group._ungrouped_tensors)
    ] + [{"string": k, "type": "GROUP"} for k in group._groups_filtered]


def _parse_last_tensor(tokens, ds):
    """Parse the last tensor name in a query. Handles tensors inside groups.

    Example:
        if "a.b.c.d" is a tensor in ds, and the query is "a.b.c.d == ", then this
        method returns ds.a.b.c.d.
    """
    tensor = []
    n = len(tokens) - 1
    if tokens[n]["type"] in ("UNKNOWN", "OP"):
        n -= 1
    for i in range(n, -1, -1):
        if tokens[i]["string"] == ".":
            continue
        elif tokens[i]["type"] in ("TENSOR", "GROUP"):
            tensor.append(tokens[i]["string"])
        else:
            break
    ret = ds
    for i in range(len(tensor) - 1, -1, -1):
        ret = ret[tensor[i]]
    return ret


def _filter(suggestions: List[dict], string: str):
    """Filter a list of suggestions. Only those suggestions that start with `string` are retained. Case insensitive."""
    return list(
        filter(lambda x: x["string"].lower().startswith(string.lower()), suggestions)
    )


def _autocomplete_response(
    suggestions: List[dict], tokens: List[dict], replace: str = ""
):
    return {
        "suggestions": suggestions,
        "tokens": tokens,
        "replace": replace,
    }


def _prefix_suggestions(suggestions: List[dict], prefix: str) -> List[dict]:
    """Prefix a list of suggestions with a string. Used to prefix suggestions with "." or " "."""
    return list(
        map(lambda s: {"string": prefix + s["string"], "type": s["type"]}, suggestions)
    )


def _op_suggestions():
    ops = [
        "==",
        ">",
        "<",
        ">=",
        "<=",
        "!=",
    ]
    return [{"string": op, "type": "OP"} for op in ops]


def autocomplete(s: str, ds: hub.Dataset) -> dict:
    """Returns autocomplete suggestions and tokenizer result for a given query string and dataset.

    Args:
        s (str): The query string
        ds (hub.Dataset): The hub dataset against which the query is to be run.

    Returns:
        dict. Auto complete response.

        Autcomplete response spec:

        {
            "suggestions": List[Suggestion],
            "tokens": List[Token],
            "replace": "..."
        }


        Suggestion:

        {
            "string": "....",
            "type": "....",
        }


        Token:

        {
            "string": "....",
            "start": 0,
            "end": 5,
            "type": "...."
        }

        Token types:

        UNKNOWN
        OP           (+, -, !, ...)
        CONSTATNT    (True, False, None)
        KEYWORD      (in, is, and, or, not, if, else, for)
        STRING
        NUMBER
        TENSOR
        GROUP
        PROPERTY     (max, min, mean, ...)
        METHOD      (count, )
    """
    if not s.strip():
        return _autocomplete_response(_initial_suggestions(ds), [])
    try:
        tokens = _parse(s, ds)
    except TokenError:
        return _autocomplete_response([], _parse_no_fail(s[:-1], ds))

    last_token = tokens[-1]
    last_type = last_token["type"]
    last_string = last_token["string"]

    if last_type == "UNKNOWN":  # incomplete query; last token is not understood
        if s[-1] == " ":
            return _autocomplete_response([], tokens)
        if len(tokens) == 1:
            suggestions = _filter(_initial_suggestions(ds), last_string)
            return _autocomplete_response(suggestions, tokens, last_string)
        if len(tokens) >= 3 and tokens[-2]["string"] == ".":
            prev_token = tokens[-3]
            prev_type = prev_token["type"]
            if prev_type == "TENSOR":
                tensor = _parse_last_tensor(tokens, ds)
                suggestions = _filter(_tensor_suggestions(ds, tensor), last_string)
                return _autocomplete_response(suggestions, tokens, last_string)
            elif prev_type == "GROUP":
                group = _parse_last_tensor(tokens, ds)
                suggestions = _filter(_group_suggestions(ds, group), last_string)
                return _autocomplete_response(suggestions, tokens, last_string)
            else:
                return _autocomplete_response([], tokens)
        suggestions = _filter(
            _initial_suggestions(ds) + _const_suggestions(), last_string
        )
        return _autocomplete_response(suggestions, tokens, last_string)
    elif last_type == "OP":
        if last_string == ".":
            if len(tokens) == 1:
                return _autocomplete_response([], tokens)
            prev_token = tokens[-2]
            prev_type = prev_token["type"]
            if prev_type == "UNKNOWN":
                return _autocomplete_response([], tokens)
            elif prev_type == "TENSOR":
                tensor = _parse_last_tensor(tokens, ds)
                return _autocomplete_response(_tensor_suggestions(ds, tensor), tokens)
            elif prev_type == "GROUP":
                group = _parse_last_tensor(tokens, ds)
                return _autocomplete_response(_tensor_suggestions(ds, group), tokens)
            else:
                return _autocomplete_response([], tokens)
        else:
            suggestions = _initial_suggestions(ds) + _const_suggestions()
            return _autocomplete_response(suggestions, tokens)
    elif last_type == "KEYWORD":
        suggestions = _initial_suggestions(ds) + _const_suggestions()
        return _autocomplete_response(
            _prefix_suggestions(suggestions, "" if s[-1] == " " else " "),
            tokens,
            last_string,
        )
    elif last_type == "TENSOR":
        tensor = _parse_last_tensor(tokens, ds)
        suggestions = _prefix_suggestions(
            _tensor_suggestions(ds, tensor), "."
        ) + _prefix_suggestions(_op_suggestions(), "" if s[-1] == " " else " ")
        return _autocomplete_response(suggestions, tokens)
    elif last_type == "GROUP":
        group = _parse_last_tensor(tokens, ds)
        suggestions = _prefix_suggestions(_group_suggestions(ds, group), ".")
        return _autocomplete_response(suggestions, tokens)
    elif last_type == "METHOD":
        return _autocomplete_response([{"string": "(", "type": "OP"}], tokens)
    elif last_type == "PROPERTY":
        suggestions = _prefix_suggestions(
            _op_suggestions(), "" if s[-1] == " " else " "
        )
        return _autocomplete_response(suggestions, tokens)
    else:
        return _autocomplete_response([], tokens)



import hub
import numpy as np


def test_commit_checkout_event(hub_cloud_ds):
    hub_cloud_ds.commit()
    hub_cloud_ds.checkout("abc", create=True)


def test_query_progress_event(hub_cloud_ds):
    with hub_cloud_ds as ds:
        ds.create_tensor("labels")
        ds.labels.append([0])
        ds.labels.append([1])
    ds.commit()
    result = ds.filter("labels == 0", progressbar=False, save_result=True)
    assert len(result) == 1


def test_compute_progress_event(hub_cloud_ds):
    with hub_cloud_ds as ds:
        ds.create_tensor("abc")

    @hub.compute
    def func(sample_in, samples_out):
        samples_out.abc.append(sample_in * np.ones((2, 2)))

    ls = list(range(10))

    func().eval(ls, hub_cloud_ds)
    for i in range(10):
        np.testing.assert_array_equal(ds.abc[i].numpy(), i * np.ones((2, 2)))


# test is empty as pytorch events aren't sent currently
def test_pytorch_progress_event(hub_cloud_ds):
    pass
